# -*- coding:utf-8 -*-
# Author: Kei Choi(hanul93@gmail.com)

# ICON Spec-1 : http://www.daubnet.com/en/file-format-ico
# ICON Spec-2 : https://formats.kaitai.io/ico/index.html

"""
Icon File Format Engine Plugin

This plugin handles ICO (Icon) format for scanning and analysis.
"""

import logging
import re
import struct

from kicomav.plugins import kavutil
from kicomav.kavcore.plugin_base import ArchivePluginBase

# Module logger
logger = logging.getLogger(__name__)

# Pattern for icon dimensions
p_name = re.compile(r"(\d+)x(\d+)")


# -------------------------------------------------------------------------
# KavMain Class
# -------------------------------------------------------------------------
class KavMain(ArchivePluginBase):
    """Icon file format handler plugin.

    This plugin provides functionality for:
    - Detecting ICO (Icon) format
    - Listing icon images within ICO files
    - Extracting individual icon images
    """

    def __init__(self):
        """Initialize the Icon plugin."""
        super().__init__(
            author="Kei Choi",
            version="1.0",
            title="Icon Engine",
            kmd_name="icon",
        )

    def __get_handle(self, filename):
        """Get or create handle for icon file.

        Args:
            filename: Path to icon file

        Returns:
            File data buffer or None
        """
        if filename in self.handle:
            return self.handle.get(filename, None)

        try:
            with open(filename, "rb") as fp:
                buf = fp.read()
            self.handle[filename] = buf
            return buf

        except (IOError, OSError) as e:
            logger.debug("Failed to open icon file %s: %s", filename, e)

        return None

    def format(self, filehandle, filename, filename_ex):
        """Analyze and detect ICO format.

        Args:
            filehandle: File data (memory mapped)
            filename: Path to file
            filename_ex: Extended filename info

        Returns:
            Dictionary with format info, or None if not recognized
        """
        ret = {}

        try:
            mm = filehandle
            if mm[:4] == b"\x00\x00\x01\x00":  # Check the header
                ret["ff_icon"] = kavutil.get_uint16(mm, 4)
                return ret

        except (IOError, OSError) as e:
            logger.debug("Format detection IO error for %s: %s", filename, e)
        except Exception as e:
            logger.warning("Unexpected error in format detection for %s: %s", filename, e)

        return None

    def arclist(self, filename, fileformat, password=None):
        """List files in the archive (icon images).

        Args:
            filename: Path to icon file
            fileformat: Format info from format() method

        Returns:
            List of [engine_id, filename] pairs
        """
        file_scan_list = []

        if "ff_icon" not in fileformat:
            return file_scan_list

        try:
            num = fileformat["ff_icon"]
            mm = self.__get_handle(filename)
            if mm is None:
                return file_scan_list

            for i in range(num):
                off = 6 + (16 * i)
                w = mm[off]
                h = 256 if mm[off + 1] == 0 else mm[off + 1]

                name = "%dx%d" % (w, h)
                file_scan_list.append(["arc_icon", name])

        except (IOError, OSError) as e:
            logger.debug("Archive list IO error for %s: %s", filename, e)
        except (IndexError, struct.error) as e:
            logger.debug("Archive list parse error for %s: %s", filename, e)
        except Exception as e:
            logger.warning("Unexpected error listing archive %s: %s", filename, e)

        return file_scan_list

    def unarc(self, arc_engine_id, arc_name, fname_in_arc):
        """Extract a file from the archive (icon image).

        Args:
            arc_engine_id: Engine ID ('arc_icon')
            arc_name: Path to icon file
            fname_in_arc: Name of icon to extract (e.g., '32x32')

        Returns:
            Extracted icon data, or None on error
        """
        if arc_engine_id != "arc_icon":
            return None

        try:
            mm = self.__get_handle(arc_name)
            if mm is None:
                return None

            num = kavutil.get_uint16(mm, 4)

            if p := p_name.search(fname_in_arc):
                fw = int(p.groups()[0])
                fh = int(p.groups()[1])

                for i in range(num):
                    off = 6 + (16 * i)
                    w = mm[off]
                    h = 256 if mm[off + 1] == 0 else mm[off + 1]

                    if w == fw and h == fh:
                        img_size = kavutil.get_uint32(mm, off + 8)
                        img_off = kavutil.get_uint32(mm, off + 12)
                        return mm[img_off : img_off + img_size]

        except (IOError, OSError) as e:
            logger.debug("Archive extract IO error for %s in %s: %s", fname_in_arc, arc_name, e)
        except Exception as e:
            logger.warning("Unexpected error extracting %s from %s: %s", fname_in_arc, arc_name, e)

        return None

    def arcclose(self):
        """Close all open archive handles."""
        for fname in list(self.handle.keys()):
            try:
                self.handle.pop(fname, None)
            except Exception as e:
                logger.debug("Archive close error for %s: %s", fname, e)



# -*- coding:utf-8 -*-
# Author: Kei Choi(hanul93@gmail.com)

"""
HWPX File Format Engine Plugin

This plugin handles HWPX (Hangul Word Processor XML) format for malware detection.
"""

import contextlib
import json
import logging
import os
import zipfile

from kicomav.plugins import kavutil
from kicomav.plugins import kernel
from kicomav.kavcore import k2security
from kicomav.kavcore.plugin_base import ArchivePluginBase

# Module logger
logger = logging.getLogger(__name__)


# -------------------------------------------------------------------------
# XML parser
# -------------------------------------------------------------------------
try:
    from defusedexpat import pyexpat as expat
except ImportError:
    from xml.parsers import expat
from xml.sax.saxutils import XMLGenerator
from xml.sax.xmlreader import AttributesImpl

try:  # pragma no cover
    from cStringIO import StringIO
except ImportError:  # pragma no cover
    try:
        from StringIO import StringIO
    except ImportError:
        from io import StringIO

from collections import OrderedDict

try:
    _basestring = basestring
except NameError:
    _basestring = str
try:
    _unicode = unicode
except NameError:
    _unicode = str


class ParsingInterrupted(Exception):
    pass


class _DictSAXHandler(object):
    def __init__(
        self,
        item_depth=0,
        item_callback=lambda *args: True,
        xml_attribs=True,
        attr_prefix="@",
        cdata_key="#text",
        force_cdata=False,
        cdata_separator="",
        postprocessor=None,
        dict_constructor=OrderedDict,
        strip_whitespace=True,
        namespace_separator=":",
        namespaces=None,
        force_list=None,
    ):
        self.path = []
        self.stack = []
        self.data = []
        self.item = None
        self.item_depth = item_depth
        self.xml_attribs = xml_attribs
        self.item_callback = item_callback
        self.attr_prefix = attr_prefix
        self.cdata_key = cdata_key
        self.force_cdata = force_cdata
        self.cdata_separator = cdata_separator
        self.postprocessor = postprocessor
        self.dict_constructor = dict_constructor
        self.strip_whitespace = strip_whitespace
        self.namespace_separator = namespace_separator
        self.namespaces = namespaces
        self.namespace_declarations = OrderedDict()  # dict is always OrderedDict
        self.force_list = force_list

    def _build_name(self, full_name):
        if not self.namespaces:
            return full_name

        i = full_name.rfind(self.namespace_separator)

        if i == -1:
            return full_name

        namespace, name = full_name[:i], full_name[i + 1 :]
        if short_namespace := self.namespaces.get(namespace, namespace):
            return self.namespace_separator.join((short_namespace, name))
        else:
            return name

    # Add attribute values to dict
    def _attrs_to_dict(self, attrs):
        if isinstance(attrs, dict):
            return attrs

        return self.dict_constructor(zip(attrs[::2], attrs[1::2]))

    # Determine whether to process the namespace for the start name
    def startNamespaceDecl(self, prefix, uri):
        self.namespace_declarations[prefix or ""] = uri

    # Process the start element
    def startElement(self, full_name, attrs):
        name = self._build_name(full_name)
        attrs = self._attrs_to_dict(attrs)

        if attrs and self.namespace_declarations:
            attrs["xmlns"] = self.namespace_declarations
            self.namespace_declarations = OrderedDict()

        self.path.append((name, attrs or None))

        if len(self.path) > self.item_depth:  # If the depth is deeper...
            self.stack.append((self.item, self.data))
            if self.xml_attribs:
                attr_entries = []
                for key, value in attrs.items():
                    key = self.attr_prefix + self._build_name(key)
                    if self.postprocessor:
                        entry = self.postprocessor(self.path, key, value)
                    else:
                        entry = (key, value)
                    if entry:
                        attr_entries.append(entry)
                attrs = self.dict_constructor(attr_entries)
            else:
                attrs = None

            self.item = attrs or None
            self.data = []

    # Process the end element
    def endElement(self, full_name):
        name = self._build_name(full_name)

        if len(self.path) == self.item_depth:
            item = self.item
            if item is None:
                item = self.cdata_separator.join(self.data) if self.data else None

            should_continue = self.item_callback(self.path, item)

            if not should_continue:
                raise ParsingInterrupted()

        if len(self.stack):
            self.process_element_end_data(name)
        else:
            self.item = None
            self.data = []
        self.path.pop()

    def process_element_end_data(self, name):
        data = self.cdata_separator.join(self.data) if self.data else None
        item = self.item
        self.item, self.data = self.stack.pop()

        if self.strip_whitespace and data:
            data = data.strip() or None

        if data and self.force_cdata and item is None:
            item = self.dict_constructor()

        if item is not None:
            if data:
                self.push_data(item, self.cdata_key, data)
            self.item = self.push_data(self.item, name, item)
        else:
            self.item = self.push_data(self.item, name, data)

    def characters(self, data):
        if not self.data:
            self.data = [data]
        else:
            self.data.append(data)

    def push_data(self, item, key, data):
        if self.postprocessor is not None:
            result = self.postprocessor(self.path, key, data)
            if result is None:
                return item
            key, data = result

        if item is None:
            item = self.dict_constructor()

        try:
            value = item[key]
            if isinstance(value, list):
                value.append(data)
            else:
                item[key] = [value, data]
        except KeyError:
            item[key] = [data] if self._should_force_list(key, data) else data
        return item

    def _should_force_list(self, key, value):
        if not self.force_list:
            return False

        if isinstance(self.force_list, bool):
            return self.force_list

        try:
            return key in self.force_list
        except TypeError:
            return self.force_list(self.path[:-1], key, value)


# Parse XML
def xml_parse(
    xml_input,
    encoding=None,
    expat=expat,
    process_namespaces=False,
    namespace_separator=":",
    disable_entities=True,
    **kwargs,
):

    handler = _DictSAXHandler(namespace_separator=namespace_separator, **kwargs)

    if isinstance(xml_input, _unicode):
        if not encoding:
            encoding = "utf-8"

        xml_input = xml_input.encode(encoding)

    if not process_namespaces:
        namespace_separator = None

    parser = expat.ParserCreate(encoding, namespace_separator)

    with contextlib.suppress(AttributeError):
        parser.ordered_attributes = True

    parser.StartNamespaceDeclHandler = handler.startNamespaceDecl
    parser.StartElementHandler = handler.startElement
    parser.EndElementHandler = handler.endElement
    parser.CharacterDataHandler = handler.characters
    parser.buffer_text = True

    if disable_entities:
        try:
            feature = "http://apache.org/xml/features/disallow-doctype-decl"
            parser._reader.setFeature(feature, True)
        except AttributeError:
            parser.DefaultHandler = lambda x: None
            parser.ExternalEntityRefHandler = lambda *x: 1

    if hasattr(xml_input, "read"):
        parser.ParseFile(xml_input)
    else:
        parser.Parse(xml_input, True)

    return handler.item


def _process_namespace(name, namespaces, ns_sep=":", attr_prefix="@"):
    if not namespaces:
        return name

    try:
        ns, name = name.rsplit(ns_sep, 1)
    except ValueError:
        pass
    else:
        ns_res = namespaces.get(ns.strip(attr_prefix))
        name = f'{attr_prefix if ns.startswith(attr_prefix) else ""}{ns_res}{ns_sep}{name}' if ns_res else name

    return name


def _emit(
    key,
    value,
    content_handler,
    attr_prefix="@",
    cdata_key="#text",
    depth=0,
    preprocessor=None,
    pretty=False,
    newl="\n",
    indent="\t",
    namespace_separator=":",
    namespaces=None,
    full_document=True,
):
    key = _process_namespace(key, namespaces, namespace_separator, attr_prefix)

    if preprocessor is not None:
        result = preprocessor(key, value)
        if result is None:
            return

        key, value = result

    if not hasattr(value, "__iter__") or isinstance(value, (_basestring, dict)):
        value = [value]

    for index, v in enumerate(value):
        if full_document and depth == 0 and index > 0:
            raise ValueError("document with multiple roots")

        if v is None:
            v = OrderedDict()
        elif isinstance(v, bool):
            v = _unicode("true") if v else _unicode("false")
        elif not isinstance(v, dict):
            v = _unicode(v)

        if isinstance(v, _basestring):
            v = OrderedDict(((cdata_key, v),))

        cdata = None
        attrs = OrderedDict()
        children = []

        for ik, iv in v.items():
            if ik == cdata_key:
                cdata = iv
                continue

            if ik.startswith(attr_prefix):
                ik = _process_namespace(ik, namespaces, namespace_separator, attr_prefix)
                if ik == "@xmlns" and isinstance(iv, dict):
                    for k, v in iv.items():
                        attr = f'xmlns{f":{k}" if k else ""}'
                        attrs[attr] = _unicode(v)
                    continue
                if not isinstance(iv, _unicode):
                    iv = _unicode(iv)
                attrs[ik[len(attr_prefix) :]] = iv
                continue

            children.append((ik, iv))

        if pretty:
            content_handler.ignorableWhitespace(depth * indent)

        content_handler.startElement(key, AttributesImpl(attrs))

        if pretty and children:
            content_handler.ignorableWhitespace(newl)

        for child_key, child_value in children:
            _emit(
                child_key,
                child_value,
                content_handler,
                attr_prefix,
                cdata_key,
                depth + 1,
                preprocessor,
                pretty,
                newl,
                indent,
                namespaces=namespaces,
                namespace_separator=namespace_separator,
            )

        if cdata is not None:
            content_handler.characters(cdata)

        if pretty and children:
            content_handler.ignorableWhitespace(depth * indent)

        content_handler.endElement(key)

        if pretty and depth:
            content_handler.ignorableWhitespace(newl)


def unparse(input_dict, output=None, encoding="utf-8", full_document=True, short_empty_elements=False, **kwargs):
    if full_document and len(input_dict) != 1:  # If the root is missing, an error occurs
        raise ValueError("Document must have exactly one root.")

    must_return = False

    if output is None:
        output = StringIO()
        must_return = True

    if short_empty_elements:
        content_handler = XMLGenerator(output, encoding, True)
    else:
        content_handler = XMLGenerator(output, encoding)

    if full_document:
        content_handler.startDocument()

    for key, value in input_dict.items():
        _emit(key, value, content_handler, full_document=full_document, **kwargs)

    if full_document:
        content_handler.endDocument()

    if must_return:
        value = output.getvalue()
        with contextlib.suppress(AttributeError):
            value = value.decode(encoding)

        return value


# -------------------------------------------------------------------------
# KavMain class
# -------------------------------------------------------------------------
class KavMain(ArchivePluginBase):
    """HWPX malware detector and archive handler plugin.

    This plugin provides functionality for:
    - Detecting HWPX exploits
    - Extracting files from HWPX archives
    - Scanning XML content for malware
    """

    def __init__(self):
        """Initialize the HWPX plugin."""
        super().__init__(
            author="Kei Choi",
            version="1.0",
            title="Hwpx Engine",
            kmd_name="hwpx",
        )

    def getinfo(self):
        """Get plugin information.

        Returns:
            Dictionary containing plugin metadata
        """
        info = super().getinfo()
        info["sig_num"] = len(self.listvirus())
        return info

    def listvirus(self):
        """Get list of detectable viruses.

        Returns:
            List of virus names
        """
        return ["Exploit.HWPX.Generic"]

    def __get_handle(self, filename):
        """Get or create handle for HWPX file.

        Args:
            filename: Path to HWPX file

        Returns:
            ZipFile object or None
        """
        if filename in self.handle:
            return self.handle.get(filename, None)

        try:
            zfile = zipfile.ZipFile(filename)
            self.handle[filename] = zfile
            return zfile

        except (IOError, OSError, zipfile.BadZipFile) as e:
            logger.debug("Failed to open HWPX file %s: %s", filename, e)

        return None

    def scan(self, filehandle, filename, fileformat, filename_ex):
        """Scan for malware.

        Args:
            filehandle: File data (memory mapped)
            filename: Path to file
            fileformat: Format info from format() method
            filename_ex: Extended filename info

        Returns:
            Tuple of (found, malware_name, malware_id, result)
        """
        zfile = None
        result = (False, "", -1, kernel.NOT_FOUND)

        try:
            if "ff_hwpx" not in fileformat:
                return result

            if self.verbose:
                print("-" * 79)
                kavutil.vprint("Engine")
                kavutil.vprint(None, "Engine", "hwpx")
                kavutil.vprint(None, "File name", os.path.split(filename)[-1])
                print()

            zfile = zipfile.ZipFile(filename)

            for name in zfile.namelist():
                if name.lower().find("mimetype") != -1:
                    data = zfile.read(name)

                    if self.verbose:
                        kavutil.vprint("mimetype")
                        kavutil.vprint(None, "body", f"{data}")
                        print()

                    if data != b"application/hwp+zip":
                        result = (True, "Exploit.HWPX.Generic", 0, kernel.INFECTED)
                        return result

                elif name.lower().find("preview/prvtext.txt") != -1:
                    pass  # PrevText.txt is not scanned

                elif name.lower().find("bindata") == -1:
                    try:
                        data = zfile.read(name)
                        if data[:5] == b"<?xml":
                            dict_data = xml_parse(data)

                            if self.verbose:
                                kavutil.vprint(name)
                                print(json.dumps(dict_data, indent=2))
                                print()
                    except Exception:
                        result = (True, "Exploit.HWPX.Generic", 0, kernel.INFECTED)
                        return result

        except (IOError, OSError) as e:
            logger.debug("Scan IO error for %s: %s", filename, e)
        except Exception as e:
            logger.warning("Unexpected error scanning %s: %s", filename, e)
        finally:
            if zfile:
                zfile.close()

        return result

    def disinfect(self, filename, malware_id):
        """Disinfect malware.

        Args:
            filename: Path to infected file
            malware_id: Malware ID to disinfect

        Returns:
            True if successful, False otherwise
        """
        try:
            if malware_id == 0:
                # CWE-73: Safe file deletion
                filename_dir = os.path.dirname(filename) or os.getcwd()
                k2security.safe_remove_file(filename, filename_dir)
                return True

        except (IOError, OSError, k2security.SecurityError) as e:
            logger.debug("Disinfect error for %s: %s", filename, e)
        except Exception as e:
            logger.warning("Unexpected error disinfecting %s: %s", filename, e)

        return False

    def arclist(self, filename, fileformat, password=None):
        """List files in the archive.

        Args:
            filename: Path to HWPX file
            fileformat: Format info from format() method

        Returns:
            List of [engine_id, filename] pairs
        """
        file_scan_list = []

        if "ff_hwpx" not in fileformat:
            return file_scan_list

        try:
            zfile = self.__get_handle(filename)
            if zfile is None:
                return file_scan_list

            # Only decompress the bindata folder data
            file_scan_list.extend(["arc_hwpx", name] for name in zfile.namelist() if name.lower().find("bindata") != -1)

        except (IOError, OSError) as e:
            logger.debug("Archive list IO error for %s: %s", filename, e)
        except Exception as e:
            logger.warning("Unexpected error listing archive %s: %s", filename, e)

        return file_scan_list

    def unarc(self, arc_engine_id, arc_name, fname_in_arc):
        """Extract a file from the archive.

        Args:
            arc_engine_id: Engine ID ('arc_hwpx')
            arc_name: Path to HWPX file
            fname_in_arc: Name of file to extract

        Returns:
            Extracted file data, or None on error
        """
        if arc_engine_id != "arc_hwpx":
            return None

        try:
            zfile = self.__get_handle(arc_name)
            if zfile is None:
                return None

            return zfile.read(fname_in_arc)

        except (IOError, OSError, zipfile.BadZipFile) as e:
            logger.debug("Archive extract error for %s in %s: %s", fname_in_arc, arc_name, e)
        except Exception as e:
            logger.warning("Unexpected error extracting %s from %s: %s", fname_in_arc, arc_name, e)

        return None

    def arcclose(self):
        """Close all open archive handles."""
        for fname in list(self.handle.keys()):
            try:
                zfile = self.handle.get(fname)
                if zfile:
                    zfile.close()
            except (IOError, OSError) as e:
                logger.debug("Archive close IO error for %s: %s", fname, e)
            except Exception as e:
                logger.debug("Archive close error for %s: %s", fname, e)
            finally:
                self.handle.pop(fname, None)



# -*- coding:utf-8 -*-
# Author: Kei Choi(hanul93@gmail.com)

"""
LNK File Format Engine Plugin

This plugin handles LNK (Windows Shortcut) format for malware detection.
"""

import contextlib
import logging
import os
import re

from kicomav.plugins import kavutil
from kicomav.plugins import kernel
from kicomav.kavcore import k2security
from kicomav.kavcore.plugin_base import MalwareDetectorBase

# Module logger
logger = logging.getLogger(__name__)


# -------------------------------------------------------------------------
# KavMain Class
# -------------------------------------------------------------------------
class KavMain(MalwareDetectorBase):
    """LNK malware detector plugin.

    This plugin provides functionality for:
    - Detecting malicious LNK files
    - Scanning for downloader patterns in shortcuts
    """

    def __init__(self):
        """Initialize the LNK plugin."""
        super().__init__(
            author="Kei Choi",
            version="1.0",
            title="LNK Scan Engine",
            kmd_name="lnk",
        )
        self.p_http = None

    def _load_virus_database(self) -> int:
        """Load virus patterns.

        Returns:
            0 for success
        """
        self.p_http = re.compile(rb"https?://")
        self.virus_names = ["Trojan.LNK.Agent.gen"]
        return 0

    def getinfo(self):
        """Get plugin information.

        Returns:
            Dictionary containing plugin metadata
        """
        info = super().getinfo()
        info["sig_num"] = 1
        return info

    def scan(self, filehandle, filename, fileformat, filename_ex):
        """Scan for malware.

        Args:
            filehandle: File data (memory mapped)
            filename: Path to file
            fileformat: Format info from format() method
            filename_ex: Extended filename info

        Returns:
            Tuple of (found, malware_name, malware_id, result)
        """
        try:
            mm = filehandle

            if mm[:8] != b"\x4c\x00\x00\x00\x01\x14\x02\x00":  # Check LNK header
                return False, "", kernel.DISINFECT_NONE, kernel.NOT_FOUND

            flag = kavutil.get_uint32(mm, 0x14)

            off = 0x4C
            if flag & 0x0001 == 0x0001:  # HasLinkTargetIDList
                clid_mycom = b"\x14\x00\x1F\x50\xE0\x4F\xD0\x20\xEA\x3A\x69\x10\xA2\xD8\x08\x00\x2B\x30\x30\x9D"
                if mm[off + 2 : off + 2 + 0x14] != clid_mycom:  # MyComputer
                    return False, "", kernel.DISINFECT_NONE, kernel.NOT_FOUND

                off += 2
                while True:
                    size = kavutil.get_uint16(mm, off)
                    if size == 0:
                        off += 2
                        break
                    if int(mm[off + 2]) == 0x32 and mm[off + 0xE : off + 0xE + 7].lower() != b"cmd.exe":
                        return False, "", kernel.DISINFECT_NONE, kernel.NOT_FOUND

                    off += size

            if flag & 0x0002 == 0x0002:  # HasLinkInfo
                off += kavutil.get_uint16(mm, off)

            if flag & 0x0004 == 0x0004:  # HasName
                size = kavutil.get_uint16(mm, off)
                off += 2 + (size * 2)

            if flag & 0x0008 == 0x0008:  # HasRelativePath
                size = kavutil.get_uint16(mm, off)
                cmd_path = mm[off + 2 : off + 2 + (size * 2) : 2].lower()

                if cmd_path.find(b"cmd.exe") == -1:
                    return False, "", kernel.DISINFECT_NONE, kernel.NOT_FOUND
                off += 2 + (size * 2)

            if flag & 0x0010 == 0x0010:  # HasWorkingDir
                size = kavutil.get_uint16(mm, off)
                off += 2 + (size * 2)

            if flag & 0x0020 == 0x0020:  # HasArguments
                size = kavutil.get_uint16(mm, off)
                cmd_arg = mm[off + 2 : off + 2 + (size * 2) : 2].lower()
                cmd_arg = cmd_arg.replace(b"^", b"")

                # Compare malware pattern
                if self.p_http and self.p_http.search(cmd_arg):
                    return True, "Trojan.LNK.Agent.gen", kernel.DISINFECT_DELETE, kernel.INFECTED

        except (IOError, OSError, ValueError) as e:
            logger.debug("Scan error for %s: %s", filename, e)
        except Exception as e:
            logger.warning("Unexpected error scanning %s: %s", filename, e)

        return False, "", kernel.DISINFECT_NONE, kernel.NOT_FOUND

    def disinfect(self, filename, malware_id):
        """Disinfect malware.

        Args:
            filename: Path to infected file
            malware_id: Malware ID to disinfect

        Returns:
            True if successful, False otherwise
        """
        try:
            if malware_id == kernel.DISINFECT_DELETE:
                filename_dir = os.path.dirname(filename) or os.getcwd()
                k2security.safe_remove_file(filename, filename_dir)
                return True

        except (IOError, OSError, k2security.SecurityError) as e:
            logger.debug("Disinfect error for %s: %s", filename, e)
        except Exception as e:
            logger.warning("Unexpected error disinfecting %s: %s", filename, e)

        return False



# -*- coding:utf-8 -*-
# Author: Kei Choi(hanul93@gmail.com)


import contextlib
import os
import re
import glob
import shutil
import tempfile


# ---------------------------------------------------------------------
# K2Tempfile Class
# ---------------------------------------------------------------------
class K2Tempfile:
    def __init__(self):
        self.re_pid = re.compile(r"ktmp([0-9a-f]{5})$", re.IGNORECASE)

        self.temp_path = os.path.join(tempfile.gettempdir(), f"ktmp{os.getpid():05x}")

        if not os.path.exists(self.temp_path):
            try:
                os.mkdir(self.temp_path)
            except (IOError, OSError) as e:
                self.temp_path = tempfile.gettempdir()

    def gettempdir(self):
        return self.temp_path

    def mktemp(self):
        """
        Create a secure temporary file and return its path.
        Uses mkstemp instead of mktemp to prevent race condition (CWE-377).
        """
        fd, path = tempfile.mkstemp(prefix="ktmp", dir=self.temp_path)
        os.close(fd)  # Close file descriptor, file remains
        return path

    def removetempdir(self):
        # Delete only the self folder
        with contextlib.suppress(OSError):
            if os.path.exists(self.temp_path):
                shutil.rmtree(self.temp_path)


# -------------------------------------------------------------------------
# FileStruct Class
# -------------------------------------------------------------------------
class FileStruct:
    # ---------------------------------------------------------------------
    # __init__(self, filename=None)
    # Initialize the class
    # Argument: filename - File name
    # ---------------------------------------------------------------------
    def __init__(self, filename=None, level=0):
        self.__fs = {}

        if filename:
            self.set_default(filename, level)

    # ---------------------------------------------------------------------
    # set_default(self, filename)
    # Create a FileStruct for a file
    # Argument: filename - File name
    # ---------------------------------------------------------------------
    def set_default(self, filename, level):
        from kicomav.plugins import kernel

        self.__fs["is_arc"] = False  # Compression status
        self.__fs["arc_engine_name"] = None  # Decompression engine ID
        self.__fs["arc_filename"] = ""  # Actual compressed file
        self.__fs["filename_in_arc"] = ""  # Decompression target file
        self.__fs["real_filename"] = filename  # Inspection target file
        self.__fs["additional_filename"] = ""  # File name for internal representation of compressed file
        self.__fs["master_filename"] = filename  # Output
        self.__fs["is_modify"] = False  # Modify status
        self.__fs["can_arc"] = kernel.MASTER_IGNORE  # Recompression possible status
        self.__fs["level"] = level  # Compression depth

    # ---------------------------------------------------------------------
    # is_archive(self)
    # Check the compression status of the file
    # Return: True or False
    # ---------------------------------------------------------------------
    def is_archive(self):  # Compression status
        return self.__fs["is_arc"]

    # ---------------------------------------------------------------------
    # get_archive_engine_name(self)
    # Check the decompression engine
    # Return: Decompression engine (ex, arc_zip)
    # ---------------------------------------------------------------------
    def get_archive_engine_name(self):  # Compression engine ID
        return self.__fs["arc_engine_name"]

    # ---------------------------------------------------------------------
    # get_archive_filename(self)
    # Check the actual compressed file name
    # Return: Actual compressed file name
    # ---------------------------------------------------------------------
    def get_archive_filename(self):  # Actual compressed file
        return self.__fs["arc_filename"]

    # ---------------------------------------------------------------------
    # get_filename_in_archive(self)
    # Check the decompression target file name
    # Return: Decompression target file
    # ---------------------------------------------------------------------
    def get_filename_in_archive(self):  # Decompression target file
        return self.__fs["filename_in_arc"]

    # ---------------------------------------------------------------------
    # get_filename(self)
    # Check the actual working target file name
    # Return: Actual working target file
    # ---------------------------------------------------------------------
    def get_filename(self):  # Actual working file name
        return self.__fs["real_filename"]

    # ---------------------------------------------------------------------
    # set_filename(self)
    # Save the actual working target file name
    # Argument: Actual working target file
    # ---------------------------------------------------------------------
    def set_filename(self, fname):  # Actual working file name
        self.__fs["real_filename"] = fname

    # ---------------------------------------------------------------------
    # get_master_filename(self)
    # Check the top file name
    # Return: Compressed file name
    # ---------------------------------------------------------------------
    def get_master_filename(self):  # Top file name
        return self.__fs["master_filename"]  # Output

    # ---------------------------------------------------------------------
    # get_additional_filename(self)
    # Check the file name for representing the compressed file
    # Return: File name for representing the compressed file
    # ---------------------------------------------------------------------
    def get_additional_filename(self):
        return self.__fs["additional_filename"]

    # ---------------------------------------------------------------------
    # set_additional_filename(self, filename)
    # Set the file name for representing the compressed file
    # ---------------------------------------------------------------------
    def set_additional_filename(self, filename):
        self.__fs["additional_filename"] = filename

    # ---------------------------------------------------------------------
    # is_modify(self)
    # Check if the file has been modified due to virus removal
    # Return: True or False
    # ---------------------------------------------------------------------
    def is_modify(self):  # Modify status
        return self.__fs["is_modify"]

    # ---------------------------------------------------------------------
    # set_modify(self, modify)
    # Save the modification status due to virus removal
    # Argument: Modify status (True or False)
    # ---------------------------------------------------------------------
    def set_modify(self, modify):  # Modify status
        self.__fs["is_modify"] = modify

    # ---------------------------------------------------------------------
    # get_can_archive(self)
    # Check if the file can be recompressed after virus removal
    # Return: kernel.MASTER_IGNORE, kernel.MASTER_PACK, kernel.MASTER_DELETE
    # ---------------------------------------------------------------------
    def get_can_archive(self):  # Recompression possible status
        return self.__fs["can_arc"]

    # ---------------------------------------------------------------------
    # set_can_archive(self, mode)
    # Set if the file can be recompressed after virus removal
    # Argument: mode - kernel.MASTER_IGNORE, kernel.MASTER_PACK, kernel.MASTER_DELETE
    # ---------------------------------------------------------------------
    def set_can_archive(self, mode):  # Recompression possible status
        self.__fs["can_arc"] = mode

    # ---------------------------------------------------------------------
    # get_level(self)
    # Check the compression depth
    # Return: 0, 1, 2 ...
    # ---------------------------------------------------------------------
    def get_level(self):  # Compression depth
        return self.__fs["level"]

    # ---------------------------------------------------------------------
    # set_level(self, level)
    # Set the compression depth
    # Argument: level - Compression depth
    # ---------------------------------------------------------------------
    def set_level(self, level):  # Compression depth
        self.__fs["level"] = level

    # ---------------------------------------------------------------------
    # set_archive(self, engine_id, rname, fname, dname, mname, modify, can_arc)
    # Save the file information given the information
    # Argument: engine_id - Decompression possible engine ID
    #           rname     - Compressed file
    #           fname     - Decompression target file
    #           dname     - File name for representing the compressed file
    #           mname     - Master file (Top file name)
    #           modify    - Modify status
    #           can_arc   - Recompression possible status
    #           level     - Compression depth
    # ---------------------------------------------------------------------
    def set_archive(self, engine_id, rname, fname, dname, mname, modify, can_arc, level):
        self.__fs["is_arc"] = True  # Compression status
        self.__fs["arc_engine_name"] = engine_id  # Decompression possible engine ID
        self.__fs["arc_filename"] = rname  # Actual compressed file
        self.__fs["filename_in_arc"] = fname  # Decompression target file
        self.__fs["real_filename"] = ""  # Inspection target file
        self.__fs["additional_filename"] = dname  # File name for representing the compressed file
        self.__fs["master_filename"] = mname  # Master file (Top file name)
        self.__fs["is_modify"] = modify  # Modify status
        self.__fs["can_arc"] = can_arc  # Recompression possible status
        self.__fs["level"] = level  # Compression depth



# -*- coding:utf-8 -*-
# Author: Kei Choi(hanul93@gmail.com)

"""
YARA Engine Plugin

This plugin handles malware detection using external YARA rules.
Supports loading multiple YARA rule files from plugins/rules/yara folder.
"""

import logging
import os
import warnings
import zipfile
from pathlib import Path

from kicomav.plugins import kernel
from kicomav.kavcore import k2security
from kicomav.kavcore.plugin_base import MalwareDetectorBase

try:
    import yara

    LOAD_YARA = True
except ImportError:
    LOAD_YARA = False

# Module logger
logger = logging.getLogger(__name__)


# -------------------------------------------------------------------------
# KavMain Class
# -------------------------------------------------------------------------
class KavMain(MalwareDetectorBase):
    """YARA-based malware detector plugin.

    This plugin provides functionality for:
    - Detecting malware using external YARA rule files
    - Loading multiple YARA rules from rules/yara folder
    - Supporting custom YARA rules with KicomAV metadata
    """

    def __init__(self):
        """Initialize the YARA Engine plugin."""
        super().__init__(
            author="Kei Choi",
            version="1.1",
            title="Yara Engine",
            kmd_name="yaraex",
        )
        self.rules = None
        self.rule_count = 0
        self.loaded_files = []
        self.failed_files = []
        self.virus_names = []  # List of virus names for listvirus()

    def _get_yara_paths(self) -> list[Path]:
        """Get YARA rules paths from rules_paths configuration.

        Uses the parent class _get_rule_path helper to find 'yara' subfolders
        in both system and user rules base paths.

        Returns:
            List of valid Path objects for YARA rules directories
        """
        return self._get_rule_path("yara")

    def _load_virus_database(self) -> int:
        """Load virus patterns from multiple YARA rules.

        Loads all .yar and .yara files from paths specified in environment variables
        (SYSTEM_RULES_BASE, USER_RULES_BASE), including rules inside ZIP archives.

        Returns:
            0 for success, -1 for failure
        """
        # If the Yara module is not installed, handle engine loading failure
        if not LOAD_YARA:
            return -1

        # Get YARA rules paths from environment variables
        rules_dirs = self._get_yara_paths()

        if not rules_dirs:
            if self.verbose:
                logger.info("No YARA rules paths configured. Set SYSTEM_RULES_BASE or USER_RULES_BASE in .env")
            return -1

        # Collect .yar and .yara files from all configured paths (including subfolders)
        rule_files = []
        for rules_dir in rules_dirs:
            rule_files.extend(rules_dir.glob("**/*.yar"))
            rule_files.extend(rules_dir.glob("**/*.yara"))

        # Build sources dict with unique namespaces
        sources = {}

        # Helper function to add source with unique namespace
        def add_source(name, source_content, source_path):
            namespace = name
            counter = 1
            original_namespace = namespace
            while namespace in sources:
                namespace = f"{original_namespace}_{counter}"
                counter += 1
            sources[namespace] = (source_content, source_path)

        # Load from regular files
        for rule_file in rule_files:
            try:
                with open(rule_file, "r", encoding="utf-8") as f:
                    source = f.read()
                add_source(rule_file.stem, source, str(rule_file))
            except Exception as e:
                self.failed_files.append((str(rule_file), str(e)))
                logger.debug("Failed to read %s: %s", rule_file, e)

        # Load from ZIP files in all configured paths
        zip_files = []
        for rules_dir in rules_dirs:
            zip_files.extend(rules_dir.glob("**/*.zip"))
        for zip_path in zip_files:
            try:
                with zipfile.ZipFile(zip_path, "r") as zf:
                    for name in zf.namelist():
                        name_lower = name.lower()
                        if name_lower.endswith(".yar") or name_lower.endswith(".yara"):
                            try:
                                source = zf.read(name).decode("utf-8")
                                rule_name = Path(name).stem
                                add_source(rule_name, source, f"{zip_path}:{name}")
                            except Exception as e:
                                self.failed_files.append((f"{zip_path}:{name}", str(e)))
                                logger.debug("Failed to read %s from %s: %s", name, zip_path, e)
            except zipfile.BadZipFile as e:
                self.failed_files.append((str(zip_path), str(e)))
                logger.debug("Bad ZIP file %s: %s", zip_path, e)
            except Exception as e:
                self.failed_files.append((str(zip_path), str(e)))
                logger.debug("Failed to open ZIP %s: %s", zip_path, e)

        if not sources:
            if self.verbose:
                logger.info("No YARA rule files found in configured paths: %s", rules_dirs)
            return -1

        # Validate each rule source individually
        # Note: Use sources instead of filepaths to avoid encoding issues with non-ASCII paths
        valid_sources = {}
        for namespace, (source, filepath) in sources.items():
            try:
                # Test compile individual source
                yara.compile(source=source)
                valid_sources[namespace] = source
                self.loaded_files.append(filepath)
            except yara.SyntaxError as e:
                self.failed_files.append((filepath, str(e)))
                logger.debug("YARA syntax error in %s: %s", filepath, e)
            except Exception as e:
                self.failed_files.append((filepath, str(e)))
                logger.debug("YARA load error in %s: %s", filepath, e)

        if not valid_sources:
            if self.verbose:
                logger.info("No valid YARA rules found")
            return -1

        # Compile all valid rules together
        try:
            self.rules = yara.compile(sources=valid_sources)

            # Count total rules and build virus names list
            self.rule_count = 0
            self.virus_names = []
            for rule in self.rules:
                self.rule_count += 1
                # Check for KicomAV metadata
                if "KicomAV" in rule.meta:
                    vname = rule.meta["KicomAV"]
                else:
                    # Fallback: use YARA.[rule_name] format
                    vname = f"YARA.{rule.identifier}"
                self.virus_names.append(vname)

            if self.verbose:
                logger.info(
                    "Loaded %d YARA rule files (%d rules), %d failed",
                    len(valid_sources),
                    self.rule_count,
                    len(self.failed_files),
                )

        except Exception as e:
            if self.verbose:
                print("[*] ERROR : YARA Rule compile")
            logger.debug("YARA rule compile error: %s", e)
            return -1

        return 0

    def getinfo(self):
        """Get plugin information.

        Returns:
            Dictionary containing plugin metadata
        """
        info = super().getinfo()
        info["sig_num"] = self.rule_count
        return info

    def listvirus(self):
        """Get list of malware that can be detected by YARA rules.

        Returns:
            List of malware names from loaded YARA rules
        """
        vlist = self.virus_names.copy()
        vlist.sort()
        return vlist

    def scan(self, filehandle, filename, fileformat, filename_ex):
        """Scan for malware using YARA rules.

        Args:
            filehandle: File data (memory mapped)
            filename: Path to file
            fileformat: Format info from format() method
            filename_ex: Extended filename info

        Returns:
            Tuple of (found, malware_name, malware_id, result)
        """
        if self.rules is None:
            return False, "", -1, kernel.NOT_FOUND

        try:
            # Exclude YARA rule files themselves
            filename_lower = filename.lower()
            if filename_lower.endswith(".yar") or filename_lower.endswith(".yara"):
                return False, "", -1, kernel.NOT_FOUND

            # Use data parameter instead of filename to avoid encoding issues
            # Filter out "too many matches" warnings from YARA
            with warnings.catch_warnings():
                warnings.filterwarnings("ignore", category=RuntimeWarning, message="too many matches")
                ret = self.rules.match(data=filehandle)
            if len(ret):
                for t in ret:
                    # Check for KicomAV metadata
                    if vname := t.meta.get("KicomAV", None):
                        return True, vname, 0, kernel.INFECTED

                    # Fallback: use rule name if no KicomAV metadata
                    # Format: YARA.[namespace].[rule_name]
                    vname = f"YARA.{t.namespace}.{t.rule}"
                    return True, vname, 0, kernel.INFECTED

        except (IOError, OSError) as e:
            logger.debug("Scan IO error for %s: %s", filename, e)
        except Exception as e:
            logger.warning("Unexpected error scanning %s: %s", filename, e)

        return False, "", -1, kernel.NOT_FOUND

    def disinfect(self, filename, malware_id):
        """Disinfect malware.

        Args:
            filename: Path to infected file
            malware_id: Malware ID to disinfect

        Returns:
            True if successful, False otherwise
        """
        try:
            if malware_id == 0:
                filename_dir = os.path.dirname(filename) or os.getcwd()
                k2security.safe_remove_file(filename, filename_dir)
                return True

        except (IOError, OSError, k2security.SecurityError) as e:
            logger.debug("Disinfect error for %s: %s", filename, e)
        except Exception as e:
            logger.warning("Unexpected error disinfecting %s: %s", filename, e)

        return False



# -*- coding:utf-8 -*-
# Author: Kei Choi(hanul93@gmail.com)

"""
UPX Unpacker Engine Plugin

This plugin handles UPX-packed PE executables for scanning.
"""

import mmap
import os
import struct
import logging

from kicomav.plugins import kernel
from kicomav.plugins import kavutil
from kicomav.plugins import pe
from kicomav.kavcore.plugin_base import ArchivePluginBase

# Module logger
logger = logging.getLogger(__name__)


# -------------------------------------------------------------------------
# UPX constants and helper data
# -------------------------------------------------------------------------
HEADERS = (
    b"\x4D\x5A\x90\x00\x02\x00\x00\x00\x04\x00\x0F\x00\xFF\xFF\x00\x00"
    b"\xB0\x00\x00\x00\x00\x00\x00\x00\x40\x00\x1A\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xD0\x00\x00\x00"
    b"\x0E\x1F\xB4\x09\xBA\x0D\x00\xCD\x21\xB4\x4C\xCD\x21\x54\x68\x69"
    b"\x73\x20\x66\x69\x6C\x65\x20\x77\x61\x73\x20\x63\x72\x65\x61\x74"
    b"\x65\x64\x20\x62\x79\x20\x4B\x69\x63\x6F\x6D\x41\x56\x2E\x20\x0D"
    b"\x0A\x43\x6F\x70\x79\x72\x69\x67\x68\x74\x20\x28\x63\x29\x20\x31"
    b"\x39\x39\x35\x2D\x32\x30\x31\x37\x20\x4B\x65\x69\x20\x43\x68\x6F"
    b"\x69\x2E\x20\x41\x6C\x6C\x20\x72\x69\x67\x68\x74\x73\x20\x72\x65"
    b"\x73\x65\x72\x76\x65\x64\x2E\x0D\x0A\x4B\x69\x63\x6F\x6D\x41\x56"
    b"\x20\x57\x65\x62\x20\x3A\x20\x68\x74\x74\x70\x3A\x2F\x2F\x77\x77"
    b"\x77\x2E\x6B\x69\x63\x6F\x6D\x61\x76\x2E\x63\x6F\x6D\x0D\x0A\x24"
)

UPX_NRV2B = b"\x11\xdb\x11\xc9\x01\xdb\x75\x07\x8b\x1e\x83\xee\xfc"
UPX_NRV2D = b"\x83\xf0\xff\x74\x78\xd1\xf8\x89\xc5\xeb\x0b\x01\xdb"
UPX_NRV2E = b"\xeb\x52\x31\xc9\x83\xe8\x03\x72\x11\xc1\xe0\x08\x8a"
UPX_LZMA1 = b"\x56\x83\xc3\x04\x53\x50\xc7\x03\x03\x00\x02\x00\x90"
UPX_LZMA2 = b"\x56\x83\xc3\x04\x53\x50\xc7\x03\x03\x00\x02\x00\x90"


# -------------------------------------------------------------------------
# UPX helper functions
# -------------------------------------------------------------------------
def PESALIGN(o, a):
    return (o // a + (o % a != 0)) * a if a else o


def ROR(x, n, bits=32):
    mask = (2**n) - 1
    mask_bits = x & mask
    return (x >> n) | (mask_bits << (bits - n))


def ROL(x, n, bits=32):
    return ROR(x, bits - n, bits)


def checkpe(dst, dsize, pehdr):
    try:
        if kavutil.get_uint32(dst, pehdr) != 0x4550:
            raise SystemError

        valign = kavutil.get_uint32(dst, pehdr + 0x38)
        if not valign:
            raise SystemError

        sectcnt = kavutil.get_uint16(dst, pehdr + 6)
        if not sectcnt:
            raise SystemError

        sections_pos = pehdr + 0xF8

        if (sections_pos + (sectcnt * 0x28)) > dsize:
            raise SystemError
    except Exception:
        sections_pos = 0
        valign = 0
        sectcnt = 0

    return sections_pos, valign, sectcnt


def ModifyCallAddr(src: bytes, dest: bytes, ep: int, upx0: int, upx1: int, baseaddr: int) -> bytes:
    try:
        ssize = len(src)
        pos = ep - upx1
        call_count = -1

        while not pos > ssize - 13:
            if (
                src[pos] == 0xE9
                and src[pos + 3] == 0xFF
                and src[pos + 4] == 0xFF
                and src[pos + 5] == 0x5E
                and src[pos + 6] == 0x89
                and src[pos + 7] == 0xF7
                and src[pos + 8] == 0xB9
            ):
                call_count = kavutil.get_uint32(src, pos + 9)
                break
            else:
                pos += 1

        if call_count == -1:
            raise SystemError

        dst = []
        pos = 0
        dcur = 0
        dsize = len(dest)

        while call_count and pos <= dsize - 5:
            if dest[pos] in {0xE9, 0xE8} and dest[pos + 1] == 0x01:
                eax = kavutil.get_uint32(dest, pos + 1)
                ax = eax & 0xFFFF
                ax >>= 8
                eax &= 0xFFFF0000
                eax |= ax
                eax = ROL(eax, 0x10)
                ah = (eax & 0xFF00) >> 8
                al = (eax & 0x00FF) << 8
                ax = ah | al
                eax &= 0xFFFF0000
                eax |= ax
                eax = uint32(eax - (baseaddr + upx0 + pos + 1))
                eax = uint32(eax + (baseaddr + upx0))
                dst += dest[dcur : pos + 1]
                dst += struct.pack("<L", eax)

                pos += 5
                dcur = pos
                call_count -= 1
            else:
                pos += 1

        dst += dest[dcur:]
        dst = bytes(dst)
    except Exception:
        return b""

    return dst


def RebuildPE(
    src: bytes, ssize: int, dst: bytes, dsize: int, ep: int, upx0: int, upx1: int, magic: list[int], dend: int
) -> bytes:
    try:
        valign = 0
        pehdr = 0
        sectcnt = 0
        sections = 0
        realstuffsz = 0
        foffset = uint32(0xD0 + 0xF8)

        if len(src) == 0 or len(dst) == 0:
            raise SystemError

        for valign in magic:
            if (
                (ep - upx1 + valign <= ssize - 5)
                and src[ep - upx1 + valign - 2] == 0x8D
                and src[ep - upx1 + valign - 1] == 0xBE
            ):
                break

        if not valign and (ep - upx1 + 0x80 < ssize - 8):
            i = 0
            while True:
                b1 = src[ep - upx1 + 0x80 + 0 + i]
                b2 = src[ep - upx1 + 0x80 + 1 + i]

                if b1 == 0x8D and b2 == 0xBE:
                    b3 = src[ep - upx1 + 0x80 + 6 + i]
                    b4 = src[ep - upx1 + 0x80 + 7 + i]

                    if b3 == 0x8B and b4 == 0x07:
                        valign = 0x80 + i + 2
                        break
                i += 1

        if valign and ISCONTAINED(0, ssize, ep - upx1 + valign, 4):
            dst_imports = struct.unpack("<l", src[ep - upx1 + valign : ep - upx1 + valign + 4])[0]

            realstuffsz = dst_imports

            if realstuffsz >= dsize:
                raise SystemError

            pehdr = dst_imports

            while (pehdr + 8 < dsize) and (struct.unpack("<l", dst[pehdr : pehdr + 4])[0]):
                pehdr += 8
                while (pehdr + 2 < dsize) and dst[pehdr]:
                    pehdr += 1
                    while (pehdr + 2 < dsize) and dst[pehdr]:
                        pehdr += 1
                    pehdr += 1
                pehdr += 1
            pehdr += 4

            sections, valign, sectcnt = checkpe(dst, dsize, pehdr)
            if not sections:
                pehdr = 0

        if not pehdr and dend > (0xF8 + 0x28):
            pehdr = dend - 0xF8 - 0x28

            while int32(pehdr) > 0:
                sections, valign, sectcnt = checkpe(dst, dsize, pehdr)
                if sections:
                    break
                pehdr -= 1

            realstuffsz = pehdr
            if not realstuffsz:
                raise SystemError

        if not pehdr:
            rebsz = uint32(PESALIGN(dend, 0x1000))
            # To Do

        foffset = PESALIGN(foffset + 0x28 * sectcnt, valign)

        for _ in range(sectcnt):
            t = kavutil.get_uint32(dst, sections + 8)
            vsize = uint32(PESALIGN(t, valign))

            t = kavutil.get_uint32(dst, sections + 12)
            urva = uint32(PESALIGN(t, valign))

            if not (upx0 + realstuffsz >= urva + vsize):
                raise SystemError

            t = struct.pack("<LLLL", vsize, urva, vsize, foffset)
            dest = dst[: sections + 8]
            dest += t
            dest += dst[sections + 24 :]
            dst = dest

            foffset += vsize
            sections += 0x28

        t = struct.pack("<L", valign)
        dest = dst[: pehdr + 0x3C]
        dest += t
        dest += dst[pehdr + 0x3C + 4 :]
        dst = dest

        newbuf = [0] * foffset

        for i in range(len(HEADERS)):
            newbuf[i] = HEADERS[i]

        for i in range(0xF8 + 0x28 * sectcnt):
            newbuf[0xD0 + i] = dst[pehdr + i]

        sections = pehdr + 0xF8

        for _ in range(sectcnt):
            t1 = kavutil.get_uint32(dst, sections + 20)
            t2 = kavutil.get_uint32(dst, sections + 12)
            t3 = kavutil.get_uint32(dst, sections + 16)

            for i in range(t3):
                newbuf[t1 + i] = dst[t2 - upx0 + i]

            sections += 0x28

        if foffset > dsize + 8192:
            raise SystemError

        dsize = foffset

        upx_d = bytes(newbuf)
    except Exception:
        return b""

    return upx_d


def int32(iv):
    if iv & 0x80000000:
        iv = -0x100000000 + iv
    return iv


def uint32(iv):
    return iv & 0xFFFFFFFF


def ISCONTAINED(bb, bb_size, sb, sb_size):
    c1 = bb_size > 0
    c2 = sb_size > 0
    c3 = sb_size <= bb_size
    c4 = sb >= bb
    c5 = (sb + sb_size) <= (bb + bb_size)
    c6 = (sb + sb_size) > bb
    c7 = sb < (bb + bb_size)

    return c1 and c2 and c3 and c4 and c5 and c6 and c7


def doubleebx(src, myebx, scur, ssize):
    oldebx = myebx

    try:
        myebx = (myebx * 2) & 0xFFFFFFFF
        if not (oldebx & 0x7FFFFFFF):
            if not ISCONTAINED(0, ssize, scur, 4):
                return -1, myebx, scur
            oldebx = kavutil.get_uint32(src, scur)
            myebx = uint32(oldebx * 2 + 1)
            scur += 4
        return (oldebx >> 31), myebx, scur
    except Exception:
        return -1, myebx, scur


def upx_inflate2b(src: bytes, dsize: int, ep: int, upx0: int, upx1: int, baseaddr: int) -> tuple[int, bytes]:
    ret = -1
    dest = ""
    myebx = 0
    scur = 0
    dcur = 0
    dst = [0] * dsize
    ssize = len(src)
    unp_offset = -1

    try:
        while True:
            while True:
                oob, myebx, scur = doubleebx(src, myebx, scur, len(src))
                if oob != 1:
                    break

                if scur >= ssize or dcur >= dsize:
                    raise SystemError

                dst[dcur] = src[scur]
                dcur += 1
                scur += 1

            if oob == -1:
                raise SystemError

            backbytes = 1

            while True:
                oob, myebx, scur = doubleebx(src, myebx, scur, len(src))
                if oob == -1:
                    raise SystemError

                backbytes = backbytes * 2 + oob

                oob, myebx, scur = doubleebx(src, myebx, scur, len(src))
                if oob == -1:
                    raise SystemError

                if oob:
                    break

            backbytes -= 3

            if backbytes >= 0:
                if scur >= ssize:
                    raise SystemError
                backbytes <<= 8
                backbytes += src[scur]
                scur += 1
                backbytes ^= 0xFFFFFFFF

                if not backbytes:
                    break
                unp_offset = int32(backbytes)

            backsize, myebx, scur = doubleebx(src, myebx, scur, len(src))
            backsize &= 0xFFFFFFFF
            if backsize == 0xFFFFFFFF:
                raise SystemError

            oob, myebx, scur = doubleebx(src, myebx, scur, len(src))
            if oob == -1:
                raise SystemError

            backsize = backsize * 2 + oob

            if not backsize:
                backsize += 1

                while True:
                    oob, myebx, scur = doubleebx(src, myebx, scur, len(src))
                    if oob == -1:
                        raise SystemError

                    backsize = backsize * 2 + oob

                    oob, myebx, scur = doubleebx(src, myebx, scur, len(src))
                    if oob != 0:
                        break

                if oob == -1:
                    raise SystemError

                backsize += 2

            if (unp_offset & 0xFFFFFFFF) < 0xFFFFF300:
                backsize += 1

            backsize += 1

            if (
                not ISCONTAINED(0, dsize, dcur + unp_offset, backsize)
                or not ISCONTAINED(0, dsize, dcur, backsize)
                or unp_offset >= 0
            ):
                raise SystemError

            for i in range(backsize):
                dst[dcur + i] = dst[dcur + unp_offset + i]

            dcur += backsize

        dest = bytes(dst)

        # Adjust call address
        dest = ModifyCallAddr(src, dest, ep, upx0, upx1, baseaddr)

        # Rebuild PE file
        magic = [0x108, 0x110, 0xD5, 0]
        dest = RebuildPE(src, ssize, dest, dsize, ep, upx0, upx1, magic, dcur)

        ret = 0
    except:
        return -1, b""

    return ret, dest


# -------------------------------------------------------------------------
# class KavMain
# -------------------------------------------------------------------------
class KavMain(ArchivePluginBase):
    """UPX unpacker plugin.

    This plugin provides functionality for:
    - Detecting UPX-packed PE executables
    - Unpacking UPX-compressed files
    """

    def __init__(self):
        """Initialize the UPX plugin."""
        super().__init__(
            author="Kei Choi",
            version="1.0",
            title="UPX Unpacker Engine",
            kmd_name="upx",
        )

    def getinfo(self):
        """Get plugin information.

        Returns:
            Dictionary containing plugin metadata
        """
        info = super().getinfo()
        info["make_arc_type"] = kernel.MASTER_DELETE
        return info

    def arclist(self, filename, fileformat, password=None):
        """List UPX-packed file.

        Args:
            filename: Path to PE file
            fileformat: Format info from format() method

        Returns:
            List of [engine_id, filename] pairs
        """
        fp = None
        mm = None
        file_scan_list = []

        if "ff_pe" not in fileformat:
            return file_scan_list

        pe_info = fileformat["ff_pe"]["pe"]
        ep_foff = pe_info["EntryPointRaw"]

        try:
            fp = open(filename, "rb")
            mm = mmap.mmap(fp.fileno(), 0, access=mmap.ACCESS_READ)

            if mm[ep_foff + 0x69 : ep_foff + 0x69 + 13] == UPX_NRV2B:
                arc_name = "arc_upx!nrv2b"
            elif mm[ep_foff + 0x71 : ep_foff + 0x71 + 13] == UPX_NRV2B:
                arc_name = "arc_upx!nrv2b"
            elif mm[ep_foff + 0x69 : ep_foff + 0x69 + 13] == UPX_NRV2D:
                arc_name = "arc_upx!nrv2d"
            elif mm[ep_foff + 0x71 : ep_foff + 0x71 + 13] == UPX_NRV2D:
                arc_name = "arc_upx!nrv2d"
            elif mm[ep_foff + 0x69 : ep_foff + 0x69 + 13] == UPX_NRV2E:
                arc_name = "arc_upx!nrv2e"
            elif mm[ep_foff + 0x71 : ep_foff + 0x71 + 13] == UPX_NRV2E:
                arc_name = "arc_upx!nrv2e"
            else:
                raise ValueError

            if self.verbose:
                self._print_upx_debug_info(filename, arc_name)

            file_scan_list.append([arc_name, "UPX"])

        except (IOError, OSError) as e:
            logger.debug("Archive list IO error for %s: %s", filename, e)
        except ValueError:
            pass  # Not UPX packed
        except Exception as e:
            logger.warning("Unexpected error listing archive %s: %s", filename, e)
        finally:
            if mm:
                mm.close()
            if fp:
                fp.close()

        return file_scan_list

    def _print_upx_debug_info(self, filename, arc_name):
        """Print UPX debug information."""
        print("-" * 79)
        kavutil.vprint("Engine")
        kavutil.vprint(None, "Engine", "upx")
        kavutil.vprint(None, "File name", os.path.split(filename)[-1])

        print()
        kavutil.vprint("UPX : only support 'nrv2b' compress method.")
        kavutil.vprint(None, "Compress Method", arc_name.split("!")[-1])
        print()

    def unarc(self, arc_engine_id, arc_name, fname_in_arc):
        """Unpack UPX-compressed file.

        Args:
            arc_engine_id: Engine ID ('arc_upx!...')
            arc_name: Path to PE file
            fname_in_arc: Name of file to extract

        Returns:
            Unpacked file data, or None on error
        """
        fp = None
        mm = None
        data = None

        if "arc_upx" not in arc_engine_id:
            return None

        try:
            fp = open(arc_name, "rb")
            mm = mmap.mmap(fp.fileno(), 0, access=mmap.ACCESS_READ)

            p = pe.PE(mm, False, arc_name)
            pe_format = p.parse()
            if pe_format is None:
                raise ValueError

            pe_img = pe_format["ImageBase"]
            pe_ep = pe_format["EntryPoint"]
            sections = pe_format["Sections"]
            ep_raw = pe_format["EntryPointRaw"]
            ep_nsec = pe_format["EntryPoint_in_Section"]

            foff = 0
            ssize = 0
            dsize = 0

            for section in sections:
                ssize = section["VirtualSize"]
                rva = section["RVA"]
                if rva <= pe_ep < rva + ssize:
                    foff = section["PointerRawData"]
                    i = sections.index(section)
                    if i != 0:
                        upx0 = sections[i - 1]["RVA"]
                        upx1 = sections[i]["RVA"]
                        dsize = sections[i - 1]["VirtualSize"] + ssize
                    break

            if ssize == 0 or dsize == 0:
                raise ValueError

            upx_data_rva = kavutil.get_uint32(mm, ep_raw + 2)
            sec_rva = sections[ep_nsec]["RVA"]

            skew = upx_data_rva - sec_rva - pe_img

            if mm[ep_raw + 1] != b"\xBE" or skew <= 0 or skew > 0xFFF:
                skew = 0
            elif skew > ssize:
                skew = 0
            else:
                raise ValueError

            data = mm[foff + skew : foff + ssize - skew]

            unpack_data = ""

            if arc_engine_id[8:] == "nrv2b":
                try:
                    ret_val, unpack_data = upx_inflate2b(data, dsize, pe_ep, upx0, upx1, pe_img)
                except OverflowError as e:
                    raise ValueError from e

            if self.verbose:
                kavutil.vprint("Decompress")
                kavutil.vprint(None, "Compressed Size", "%d" % len(data))
                if unpack_data == "":
                    kavutil.vprint(None, "Decompress Size", "Error")
                else:
                    kavutil.vprint(None, "Decompress Size", "%d" % len(unpack_data))
                print()

            if unpack_data == "":
                raise ValueError

            data = unpack_data

        except (IOError, OSError) as e:
            logger.debug("Archive extract IO error for %s in %s: %s", fname_in_arc, arc_name, e)
            data = None
        except ValueError:
            data = None
        except Exception as e:
            logger.warning("Unexpected error extracting %s from %s: %s", fname_in_arc, arc_name, e)
            data = None
        finally:
            if mm:
                mm.close()
            if fp:
                fp.close()

        return data



# -*- coding:utf-8 -*-
# Author: Kei Choi(hanul93@gmail.com)

# ref: https://github.com/lifenjoiner/ISx

"""
PE File Format Engine Plugin

This plugin handles PE (Portable Executable) format for scanning and analysis.
It also provides resource extraction capabilities.
"""

import binascii
import contextlib
import ctypes
import logging
import os
import re
import struct
import zlib

from kicomav.plugins import cryptolib
from kicomav.plugins import kavutil
from kicomav.plugins import kernel
from kicomav.kavcore.plugin_base import ArchivePluginBase

# Module logger
logger = logging.getLogger(__name__)

BYTE = ctypes.c_ubyte
WORD = ctypes.c_ushort
DWORD = ctypes.c_uint
FLOAT = ctypes.c_float
LPBYTE = ctypes.POINTER(ctypes.c_ubyte)
LPTSTR = ctypes.POINTER(ctypes.c_char)
HANDLE = ctypes.c_void_p
PVOID = ctypes.c_void_p
LPVOID = ctypes.c_void_p
UINT_PTR = ctypes.c_uint
SIZE_T = ctypes.c_uint


class DOS_HEADER(ctypes.Structure):
    _pack_ = 1
    _fields_ = [
        ("e_magic", WORD),
        ("e_cblp", WORD),
        ("e_cp", WORD),
        ("e_crlc", WORD),
        ("e_cparhdr", WORD),
        ("e_minalloc", WORD),
        ("e_maxalloc", WORD),
        ("e_ss", WORD),
        ("e_sp", WORD),
        ("e_csum", WORD),
        ("e_ip", WORD),
        ("e_cs", WORD),
        ("e_lfarlc", WORD),
        ("e_ovno", WORD),
        ("e_res", BYTE * 8),  # 8Byte
        ("e_oemid", WORD),
        ("e_oeminfo", WORD),
        ("e_res2", BYTE * 20),  # 20Byte
        ("e_lfanew", DWORD),
    ]


class FILE_HEADER(ctypes.Structure):
    _pack_ = 1
    _fields_ = [
        ("Machine", WORD),
        ("NumberOfSections", WORD),
        ("CreationYear", DWORD),
        ("PointerToSymbolTable", DWORD),
        ("NumberOfSymbols", DWORD),
        ("SizeOfOptionalHeader", WORD),
        ("Characteristics", WORD),
    ]


class OPTIONAL_HEADER(ctypes.Structure):
    _pack_ = 1
    _fields_ = [
        ("Magic", WORD),
        ("MajorLinkerVersion", BYTE),
        ("MinorLinkerVersion", BYTE),
        ("SizeOfCode", DWORD),
        ("SizeOfInitializedData", DWORD),
        ("SizeOfUninitializedData", DWORD),
        ("AddressOfEntryPoint", DWORD),
        ("BaseOfCode", DWORD),
        ("BaseOfData", DWORD),
        ("ImageBase", DWORD),
        ("SectionAlignment", DWORD),
        ("FileAlignment", DWORD),
        ("MajorOperatingSystemVersion", WORD),
        ("MinorOperatingSystemVersion", WORD),
        ("MajorImageVersion", WORD),
        ("MinorImageVersion", WORD),
        ("MajorSubsystemVersion", WORD),
        ("MinorSubsystemVersion", WORD),
        ("Reserved1", DWORD),
        ("SizeOfImage", DWORD),
        ("SizeOfHeaders", DWORD),
        ("CheckSum", DWORD),
        ("Subsystem", WORD),
        ("DllCharacteristics", WORD),
        ("SizeOfStackReserve", DWORD),
        ("SizeOfStackCommit", DWORD),
        ("SizeOfHeapReserve", DWORD),
        ("SizeOfHeapCommit", DWORD),
        ("LoaderFlags", DWORD),
        ("NumberOfRvaAndSizes", DWORD),
    ]


class DATA_DIRECTORY(ctypes.Structure):
    _pack_ = 1
    _fields_ = [
        ("VirtualAddress", DWORD),
        ("Size", DWORD),
    ]


class SECTION_HEADER(ctypes.Structure):
    _pack_ = 1
    _fields_ = [
        ("Name", BYTE * 8),
        ("Misc_VirtualSize", DWORD),
        ("VirtualAddress", DWORD),
        ("SizeOfRawData", DWORD),
        ("PointerToRawData", DWORD),
        ("PointerToRelocations", DWORD),
        ("PointerToLinenumbers", DWORD),
        ("NumberOfRelocations", WORD),
        ("NumberOfLinenumbers", WORD),
        ("Characteristics", DWORD),
    ]


def enum(*sequential, **named):
    enums = dict(zip(sequential, range(len(sequential))), **named)
    reverse = {value: key for key, value in enums.items()}
    enums["reverse_mapping"] = reverse
    return type("Enum", (), enums)


image_directory_entry = enum(
    "EXPORT",
    "IMPORT",
    "RESOURCE",
    "EXCEPTION",
    "SECURITY",
    "BASERELOC",
    "DEBUG",
    "COPYRIGHT",  # Architecture on non-x86 platforms
    "GLOBALPTR",
    "TLS",
    "LOAD_CONFIG",
    "BOUND_IMPORT",
    "IAT",
    "DELAY_IMPORT",
    "COM_DESCRIPTOR",
    "RESERVED",
)


p_str = re.compile(rb"[^\x00]*")  # Copy up to the NULL character


class PE:
    def __init__(self, mm, verbose, filename):
        self.filename = filename
        self.filesize = os.path.getsize(filename)
        self.verbose = verbose
        self.mm = mm
        self.sections = []  # List to hold all section information
        self.data_directories = []  # List to hold all data directory information
        self.pe_file_align = 0

    # -------------------------------------------------------------------------
    # parse(mm)
    # Parse the PE file and return the main information.
    # Input : mm - File handle
    # Return : {PE file analysis information} or None
    # -------------------------------------------------------------------------
    def parse(self):
        pe_format = {
            "PE_Position": 0,
            "EntryPoint": 0,
            "SectionNumber": 0,
            "Sections": None,
            "EntryPointRaw": 0,
            "FileAlignment": 0,
        }

        mm = self.mm
        with contextlib.suppress(ValueError, struct.error):
            self.parse_pe_parse(mm, pe_format)
            return pe_format

        return None

    def parse_pe_parse(self, mm, pe_format):
        if mm[:2] != b"MZ":  # Does it start with MZ?
            raise ValueError

        dos_header = DOS_HEADER()
        ctypes.memmove(ctypes.addressof(dos_header), mm[:], ctypes.sizeof(dos_header))

        # Get the position of the PE signature
        pe_pos = dos_header.e_lfanew

        # Is it a PE file?
        if mm[pe_pos : pe_pos + 4] != b"PE\x00\x00":
            raise ValueError

        pe_format["PE_Position"] = pe_pos

        # Read the File Header
        file_header = FILE_HEADER()
        file_header_size = ctypes.sizeof(file_header)  # file_header_size : 0x14
        ctypes.memmove(ctypes.addressof(file_header), mm[pe_pos + 4 :], file_header_size)

        # Read the Optional Header
        optional_header = OPTIONAL_HEADER()
        optional_header_size = ctypes.sizeof(optional_header)
        ctypes.memmove(
            ctypes.addressof(optional_header),
            mm[pe_pos + 4 + file_header_size :],
            optional_header_size,
        )

        # Is the Magic ID of the Optional Header correct?
        if optional_header.Magic != 0x10B:
            raise ValueError

        # Get the Entry Point
        pe_ep = optional_header.AddressOfEntryPoint
        pe_format["EntryPoint"] = pe_ep

        # Get the Image Base
        pe_img = optional_header.ImageBase
        pe_format["ImageBase"] = pe_img

        # Get the File Alignment
        self.pe_file_align = optional_header.FileAlignment
        pe_format["FileAlignment"] = self.pe_file_align

        # Get the number of sections
        section_num = file_header.NumberOfSections
        pe_format["SectionNumber"] = section_num

        # Get the size of the Optional Header
        opthdr_size = file_header.SizeOfOptionalHeader
        pe_format["OptionalHederSize"] = opthdr_size

        # Read the Data Directory
        data_directory_size = ctypes.sizeof(DATA_DIRECTORY())  # data_directory_size : 8
        num_data_directory = (opthdr_size - optional_header_size) // data_directory_size
        off_data_directory = pe_pos + 4 + file_header_size + optional_header_size

        for i in range(num_data_directory):
            dx = DATA_DIRECTORY()
            ctypes.memmove(
                ctypes.addressof(dx),
                mm[off_data_directory + (i * data_directory_size) :],
                data_directory_size,
            )

            self.data_directories.append(dx)

        # Start position of the section
        section_pos = pe_pos + 4 + file_header_size + opthdr_size

        # Extract all section information
        for i in range(section_num):
            section_header = SECTION_HEADER()
            section_header_size = ctypes.sizeof(section_header)  # section_header_size : 0x28

            s = section_pos + (section_header_size * i)
            ctypes.memmove(ctypes.addressof(section_header), mm[s:], section_header_size)

            sec_name = ctypes.cast(section_header.Name, ctypes.c_char_p)
            section = {
                "Name": sec_name.value,
                "VirtualSize": section_header.Misc_VirtualSize,
                "RVA": section_header.VirtualAddress,
                "SizeRawData": section_header.SizeOfRawData,
                "PointerRawData": section_header.PointerToRawData,
                "Characteristics": section_header.Characteristics,
            }
            self.sections.append(section)

        pe_format["Sections"] = self.sections

        # Get the position of the EntryPoint in the file
        ep_raw, sec_idx = self.rva_to_off(pe_ep)
        pe_format["EntryPointRaw"] = ep_raw  # Raw position of the EP
        pe_format["EntryPoint_in_Section"] = sec_idx  # Section containing the EP

        # Analyze resources
        try:
            rsrc_rva = self.data_directories[image_directory_entry.RESOURCE].VirtualAddress
            rsrc_size = self.data_directories[image_directory_entry.RESOURCE].Size
        except IndexError:
            rsrc_rva = 0
            rsrc_size = 0

        if rsrc_rva:  # Does the resource exist?
            # with contextlib.suppress(struct.error, ValueError):
            self.parse_pe_resources(rsrc_rva, mm, rsrc_size, pe_format)

        # Analyze import API
        try:
            imp_rva = self.data_directories[image_directory_entry.IMPORT].VirtualAddress  # Import API location (RVA)
            imp_size = self.data_directories[image_directory_entry.IMPORT].Size  # Import API size
        except IndexError:
            imp_rva = 0
            imp_size = 0

        if imp_rva:  # Import API exists
            imp_api = {}

            imp_off = self.rva_to_off(imp_rva)[0]
            imp_data = mm[imp_off : imp_off + imp_size]

            if len(imp_data) == imp_size:
                for i in range(imp_size // 0x14):  # DLL information size is 0x14
                    with contextlib.suppress(struct.error):
                        dll_rva = kavutil.get_uint32(imp_data, (i * 0x14) + 0xC)
                        api_rva = kavutil.get_uint32(imp_data, (i * 0x14))
                        bo = 2
                        if api_rva == 0:
                            api_rva = kavutil.get_uint32(imp_data, (i * 0x14) + 0x10)
                            bo = 0

                        # print (hex(api_rva))
                        if dll_rva == 0:  # DLL information is missing
                            break

                        t_off = self.rva_to_off(dll_rva)[0]
                        dll_name = p_str.search(mm[t_off : t_off + 0x20]).group()
                        # print ('[+]', dll_name)
                        imp_api[dll_name] = []

                        t_off = self.rva_to_off(api_rva)[0]
                        while True:
                            try:
                                api_name_rva = kavutil.get_uint32(mm, t_off)
                            except struct.error:
                                break

                            if api_name_rva & 0x80000000 == 0x80000000:  # Odinal API
                                t_off += 4
                                continue

                            if api_name_rva == 0:
                                break

                            t = self.rva_to_off(api_name_rva)[0]
                            # print (hex(t_off), hex(t))
                            api_name = p_str.search(mm[t + bo : t + bo + 0x20]).group()
                            # print ('   ', api_name)
                            imp_api[dll_name].append(api_name)
                            t_off += 4
            # end if

            pe_format["Import_API"] = imp_api

        # Analyze digital certificate
        try:
            cert_off = self.data_directories[
                image_directory_entry.SECURITY
            ].VirtualAddress  # The only offset that is not RVA
            cert_size = self.data_directories[image_directory_entry.SECURITY].Size  # Digital certificate size
        except IndexError:
            cert_off = 0
            cert_size = 0

        # Digital certificate exists and UPack's case, an unusual value is set
        if cert_off and cert_off + cert_size <= len(mm[:]):
            pe_format["CERTIFICATE_Offset"] = cert_off
            pe_format["CERTIFICATE_Size"] = cert_size

        # Analyze debug information
        try:
            debug_rva = self.data_directories[image_directory_entry.DEBUG].VirtualAddress  # RVA
            debug_size = self.data_directories[image_directory_entry.DEBUG].Size  # Size
            if debug_size < 0x1C:
                raise ValueError
        except (IndexError, ValueError) as e:
            debug_rva = 0
            debug_size = 0

        if debug_rva:  # Debug information exists
            t = self.rva_to_off(debug_rva)[0]
            debug_off = kavutil.get_uint32(mm, t + 0x18)
            debug_size = kavutil.get_uint32(mm, t + 0x10)

            debug_data = mm[debug_off : debug_off + debug_size]

            if debug_data[:4] == b"RSDS":
                pe_format["PDB_Name"] = debug_data[0x18:]
            else:
                pe_format["PDB_Name"] = b"Not support Type : " + debug_data[:4]

        if self.verbose:
            self.print_pe_debug_info(mm, pe_format, section_num)

    # Print PE debug information
    def print_pe_debug_info(self, mm, pe_format, section_num):
        print("-" * 79)
        kavutil.vprint("Engine")
        kavutil.vprint(None, "Engine", "pe")
        kavutil.vprint(None, "File name", os.path.split(self.filename)[-1])
        kavutil.vprint(None, "MD5", cryptolib.md5(mm[:]))

        print()
        kavutil.vprint("PE")
        kavutil.vprint(None, "EntryPoint", "%08X" % pe_format["EntryPoint"])
        kavutil.vprint(
            None,
            "EntryPoint (Section)",
            "%d" % pe_format["EntryPoint_in_Section"],
        )

        # View sections
        if section_num:
            print()
            kavutil.vprint("Section Header")
            print("    %-8s %-8s %-8s %-8s %-8s %-8s" % ("Name", "VOFF", "VSIZE", "FOFF", "FSIZE", "EXEC"))
            print("    " + ("-" * (9 * 6 - 1)))

            for s in self.sections:
                print(
                    "    %-8s %08X %08X %08X %08X %-05s"
                    % (
                        s["Name"],
                        s["RVA"],
                        s["VirtualSize"],
                        s["PointerRawData"],
                        s["SizeRawData"],
                        s["Characteristics"] & 0x20000000 == 0x20000000,
                    )
                )

        if section_num:
            print()
            kavutil.vprint("Section MD5")
            print("    %-8s %-8s %-32s" % ("Name", "FSIZE", "MD5"))
            print("    " + ("-" * ((9 * 2 - 1) + 32)))

            for s in self.sections:
                off = s["PointerRawData"]
                size = s["SizeRawData"]
                fmd5 = cryptolib.md5(mm[off : off + size]) if size else "-"
                print("    %-8s %8d %s" % (s["Name"], size, fmd5))

        print()
        kavutil.vprint("Entry Point (Raw)")
        print()
        kavutil.HexDump().Buffer(mm[:], pe_format["EntryPointRaw"], 0x80)
        print()
        if "PDB_Name" in pe_format:
            kavutil.vprint("PDB Information")
            kavutil.vprint(None, "Name", f'{repr(pe_format["PDB_Name"])}')
            print(repr(pe_format["PDB_Name"]))
            print()

    def parse_pe_resources(self, rsrc_rva, mm, rsrc_size, pe_format):
        rsrc_off, rsrc_idx = self.rva_to_off(rsrc_rva)  # Convert resource position

        if rsrc_off > self.filesize:
            raise ValueError

        t_size = self.sections[rsrc_idx]["SizeRawData"]
        if len(mm[rsrc_off : rsrc_off + rsrc_size]) != rsrc_size and len(mm[rsrc_off : rsrc_off + t_size]) != t_size:
            raise ValueError

        # Check the type
        num_type_name = kavutil.get_uint16(mm, rsrc_off + 0xC)
        num_type_id = kavutil.get_uint16(mm, rsrc_off + 0xE)

        for i in range(num_type_name + num_type_id):
            type_id = kavutil.get_uint32(mm, rsrc_off + 0x10 + (i * 8))
            name_id_off = kavutil.get_uint32(mm, rsrc_off + 0x14 + (i * 8))

            # Type is user-defined name or RCDATA?
            if type_id & 0x80000000 == 0x80000000 or type_id == 0xA or type_id == 0:
                if type_id & 0x80000000 == 0x80000000:
                    # Extract the user-defined name
                    string_off = (type_id & 0x7FFFFFFF) + rsrc_off
                    len_name = kavutil.get_uint16(mm, string_off)
                    rsrc_type_name = mm[string_off + 2 : string_off + 2 + (len_name * 2) : 2].decode("utf-8")
                elif type_id == 0xA:
                    rsrc_type_name = "RCDATA"
                else:
                    rsrc_type_name = f"{type_id}"

                # Name ID
                name_id_off = (name_id_off & 0x7FFFFFFF) + rsrc_off
                if name_id_off > self.filesize:
                    raise ValueError

                num_name_id_name = kavutil.get_uint16(mm, name_id_off + 0xC)
                num_name_id_id = kavutil.get_uint16(mm, name_id_off + 0xE)

                for j in range(num_name_id_name + num_name_id_id):
                    name_id_id = kavutil.get_uint32(mm, name_id_off + 0x10 + (j * 8))
                    language_off = kavutil.get_uint32(mm, name_id_off + 0x14 + (j * 8))

                    if name_id_id & 0x80000000 == 0x80000000:
                        string_off = (name_id_id & 0x7FFFFFFF) + rsrc_off
                        if string_off > self.filesize:
                            raise ValueError

                        len_name = kavutil.get_uint16(mm, string_off)
                        rsrc_name_id_name = mm[string_off + 2 : string_off + 2 + (len_name * 2) : 2]
                        string_name = f"{rsrc_type_name}/{rsrc_name_id_name.decode('utf-8')}"
                    else:
                        string_name = f"{rsrc_type_name}/{hex(name_id_id).upper()[2:]}"

                    # Language
                    language_off = (language_off & 0x7FFFFFFF) + rsrc_off
                    if language_off > self.filesize:
                        raise ValueError

                    num_language_name = kavutil.get_uint16(mm, language_off + 0xC)
                    num_language_id = kavutil.get_uint16(mm, language_off + 0xE)

                    for k in range(num_language_name + num_language_id):
                        data_entry_off = kavutil.get_uint32(mm, language_off + 0x14 + (k * 8))
                        data_entry_off = (data_entry_off & 0x7FFFFFFF) + rsrc_off

                        data_rva = kavutil.get_uint32(mm, data_entry_off)
                        data_off, _ = self.rva_to_off(data_rva)
                        if data_off > self.filesize:
                            continue

                        data_size = kavutil.get_uint32(mm, data_entry_off + 4)
                        if data_size > self.filesize:
                            continue

                        if data_size > 8192:  # Extract only resources of at least 8K
                            if "Resource_UserData" in pe_format:
                                pe_format["Resource_UserData"][string_name] = (data_off, data_size)
                            else:
                                pe_format["Resource_UserData"] = {string_name: (data_off, data_size)}

    def rva_to_off(self, t_rva):
        for section in self.sections:
            size = section["SizeRawData"]
            rva = section["RVA"]

            if rva <= t_rva < rva + size:
                if self.pe_file_align:
                    foff = (section["PointerRawData"] // self.pe_file_align) * self.pe_file_align
                else:
                    foff = section["PointerRawData"]
                t_off = t_rva - rva + foff

                return t_off, self.sections.index(section)

        return t_rva, -1  # If not associated with any section, return RVA


# -------------------------------------------------------------------------
# KavMain class
# -------------------------------------------------------------------------
class KavMain(ArchivePluginBase):
    """PE file format handler plugin.

    This plugin provides functionality for:
    - Detecting PE (Portable Executable) format
    - Parsing PE headers, sections, and resources
    - Detecting NSIS, InstallShield installers
    - Extracting resources from PE files
    """

    def __init__(self):
        """Initialize the PE plugin."""
        super().__init__(
            author="Kei Choi",
            version="1.2",
            title="PE Engine",
            kmd_name="pe",
        )
        self.p_nsis = None

    def _custom_init(self) -> int:
        """Custom initialization for PE plugin.

        Returns:
            0 for success
        """
        # NSIS code pattern
        # 81 7D DC EF BE AD DE    cmp     [ebp+var_24], 0DEADBEEFh
        # 75 69                   jnz     short loc_402D79
        # 81 7D E8 49 6E 73 74    cmp     [ebp+var_18], 'tsnI'
        # ...
        self.p_nsis = re.compile(
            r"81 7D .. EF BE AD DE 75 .. 81 7D .. 49 6E 73 74 75 .. 81 7D .. 73 6F 66 74 75 .. 81 7D .. 4E 75 6C 6C 75".replace(
                " ", ""
            ),
            re.IGNORECASE,
        )
        return 0

    def getinfo(self):
        """Get plugin information.

        Returns:
            Dictionary containing plugin metadata
        """
        info = super().getinfo()
        info["make_arc_type"] = kernel.MASTER_DELETE
        return info

    def format(self, filehandle, filename, filename_ex):
        """Analyze and detect PE format.

        Args:
            filehandle: File data (memory mapped)
            filename: Path to file
            filename_ex: Extended filename info

        Returns:
            Dictionary with format info, or None if not recognized
        """
        ret = {}

        try:
            pe = PE(filehandle, self.verbose, filename)
            try:
                pe_format = pe.parse()
            except MemoryError:
                pe_format = None

            if pe_format is None:
                return None

            fileformat = {"pe": pe_format}
            ret = {"ff_pe": fileformat}

            # Check if there is additional information at the end of the PE file
            pe_size = 0
            last_section_off = 0

            pe_file_align = pe_format["FileAlignment"]

            for sec in pe_format.get("Sections", []):
                if pe_file_align:
                    off = (sec["PointerRawData"] // pe_file_align) * pe_file_align
                else:
                    off = sec["PointerRawData"]
                size = sec["SizeRawData"]
                pe_size = max(pe_size, off + size)
                last_section_off = pe_size

            file_size = len(filehandle)

            if "CERTIFICATE_Offset" in pe_format:
                if pe_format["CERTIFICATE_Offset"] == pe_size:
                    t_pe_size = pe_format["CERTIFICATE_Offset"] + pe_format["CERTIFICATE_Size"]
                    pe_size = max(pe_size, t_pe_size)
                    attach_size = file_size - pe_size
                else:
                    attach_size = file_size - pe_size - pe_format["CERTIFICATE_Size"]
            else:
                attach_size = file_size - pe_size

            if pe_size < file_size and pe_size != 0:
                mm = filehandle
                self._detect_installers(mm, pe_format, pe_file_align, pe_size, last_section_off, attach_size, ret)

        except (IOError, OSError) as e:
            logger.debug("Format detection IO error for %s: %s", filename, e)
        except Exception as e:
            logger.warning("Unexpected error in format detection for %s: %s", filename, e)

        return ret if ret else None

    def _detect_installers(self, mm, pe_format, pe_file_align, pe_size, last_section_off, attach_size, ret):
        """Detect various installer formats in PE file.

        Args:
            mm: File data
            pe_format: Parsed PE format info
            pe_file_align: PE file alignment
            pe_size: PE file size
            last_section_off: Last section offset
            attach_size: Attached data size
            ret: Result dictionary to update
        """
        # Check if NSIS code exists in the .text section
        text_sec = pe_format["Sections"][0]
        if pe_file_align:
            off = (text_sec["PointerRawData"] // pe_file_align) * pe_file_align
        else:
            off = text_sec["PointerRawData"]
        size = text_sec["SizeRawData"]

        # Process NSIS
        hex_data = binascii.hexlify(mm[off : off + size]).decode()
        if self.p_nsis and self.p_nsis.search(hex_data):
            i = 1
            while True:
                t = mm[i * 0x200 + 4 : i * 0x200 + 20]
                if len(t) != 16:
                    break

                if t == b"\xEF\xBE\xAD\xDENullsoftInst":
                    ret["ff_nsis"] = {"Offset": i * 0x200}
                    break

                i += 1

        # Process InstallShield
        ishield_header = mm[-0x30:]
        if ishield_header[0x18:0x1C] == b"\x01\x91\x48\x48":
            if mm[last_section_off : last_section_off + 14] == b"InstallShield\x00":
                version = "2.0"  # INI file based InstallShield
            else:
                version = "x.x"  # CAB based InstallShield

            ret["ff_installshield"] = {
                "Header": ishield_header,
                "Attached_Pos": last_section_off,
                "Version": version,
            }

        # Process InstallShield for InstallStream
        if mm[last_section_off : last_section_off + 14] == b"ISSetupStream\x00":
            ret["ff_installshield_setup_stream"] = {
                "Attached_Pos": last_section_off,
            }

        # Process InstallShield embedded setup.exe
        ishield_embedded_file_header = mm[-712:]
        decrypted = cryptolib.ishield_decrypt(ishield_embedded_file_header, b"%eR@toPm|<#YKs$^")
        v1 = kavutil.get_uint32(decrypted, 668)
        v2 = kavutil.get_uint32(decrypted, 12)
        file_name_size = kavutil.get_uint32(decrypted, 144)
        crc32_value = kavutil.get_uint32(decrypted, 704)

        if v1 <= 0x20 and v2 <= 0x40 and file_name_size <= 0x104:
            decrypted = cryptolib.reset_padding(decrypted, 668, v1, 0x20)
            decrypted = cryptolib.reset_padding(decrypted, 12, v2, 0x40)
            decrypted = cryptolib.reset_padding(decrypted, 144, file_name_size, 0x104)
            crc32_value_calculated = int(zlib.crc32(decrypted[:704]) & 0xFFFFFFFF)

            if crc32_value == crc32_value_calculated:
                ret["ff_installshield_embedded_file"] = {
                    "Embedded_File_Pos": pe_size,
                    "Embedded_File_Size": file_name_size,
                }

        # Process attachments (do not process if NSIS or InstallShield exists)
        if (
            "ff_nsis" not in ret
            and "ff_installshield" not in ret
            and "ff_installshield_embedded_file" not in ret
            and "ff_installshield_setup_stream" not in ret
        ):
            fileformat = {
                "Attached_Pos": pe_size,
                "Attached_Size": attach_size,
            }
            ret["ff_attach"] = fileformat

    def arclist(self, filename, fileformat, password=None):
        """List files in the archive (PE resources).

        Args:
            filename: Path to PE file
            fileformat: Format info from format() method

        Returns:
            List of [engine_id, filename] pairs
        """
        file_scan_list = []

        try:
            if "ff_pe" in fileformat and "Resource_UserData" in fileformat["ff_pe"]["pe"]:
                for key in fileformat["ff_pe"]["pe"]["Resource_UserData"].keys():
                    off, size = fileformat["ff_pe"]["pe"]["Resource_UserData"][key]
                    file_scan_list.append(["arc_pe_rcdata:%d:%d" % (off, size), key])

        except (KeyError, TypeError) as e:
            logger.debug("Archive list error for %s: %s", filename, e)
        except Exception as e:
            logger.warning("Unexpected error listing resources in %s: %s", filename, e)

        return file_scan_list

    def unarc(self, arc_engine_id, arc_name, fname_in_arc):
        """Extract a file from the archive (PE resource).

        Args:
            arc_engine_id: Engine ID (format: 'arc_pe_rcdata:offset:size')
            arc_name: Path to PE file
            fname_in_arc: Name of resource to extract

        Returns:
            Extracted file data, or None on error
        """
        if arc_engine_id.find("arc_pe_rcdata:") == -1:
            return None

        try:
            t = arc_engine_id.split(":")
            off = int(t[1])
            size = int(t[2])

            with open(arc_name, "rb") as fp:
                fp.seek(off)
                data = fp.read(size)

            return data

        except (IOError, OSError) as e:
            logger.debug("Resource extract IO error for %s in %s: %s", fname_in_arc, arc_name, e)
        except (ValueError, IndexError) as e:
            logger.debug("Resource extract parse error: %s", e)
        except Exception as e:
            logger.warning("Unexpected error extracting resource %s from %s: %s", fname_in_arc, arc_name, e)

        return None

    def arcclose(self):
        """Close all open archive handles."""
        pass  # No persistent handles to close



# -*- coding:utf-8 -*-
# Author: Kei Choi(hanul93@gmail.com)

"""
Zip Archive Engine Plugin

This plugin handles ZIP and 7Z archive formats for scanning and manipulation.
"""

import os
import shutil
import zipfile
import contextlib
import logging
import tempfile

import py7zr

from kicomav.plugins import kavutil
from kicomav.plugins import kernel
from kicomav.kavcore import k2security
from kicomav.kavcore.plugin_base import ArchivePluginBase

# Module logger
logger = logging.getLogger(__name__)


# -------------------------------------------------------------------------
# class KavMain
# -------------------------------------------------------------------------
class KavMain(ArchivePluginBase):
    """ZIP and 7Z archive handler plugin.

    This plugin provides functionality for:
    - Detecting ZIP and 7Z archive formats
    - Listing files within archives
    - Extracting files from archives
    - Creating/updating archives
    """

    def __init__(self):
        """Initialize the ZIP plugin."""
        super().__init__(
            author="Kei Choi",
            version="1.1",
            title="Zip Archive Engine",
            kmd_name="zip",
        )
        # self.handle = {} is initialized in ArchivePluginBase
        self.password = None  # Password for encrypted archives

    def _custom_init(self) -> int:
        """Custom initialization for ZIP plugin.

        Returns:
            0 for success
        """
        # self.handle = {} is already initialized in ArchivePluginBase
        return 0

    def _custom_uninit(self) -> int:
        """Custom cleanup for ZIP plugin.

        Returns:
            0 for success
        """
        self.arcclose()
        return 0

    def getinfo(self):
        """Get plugin information.

        Returns:
            Dictionary containing plugin metadata
        """
        info = super().getinfo()
        info["engine_type"] = kernel.ARCHIVE_ENGINE
        info["make_arc_type"] = kernel.MASTER_PACK
        return info

    def __get_handle(self, filename):
        """Get or create handle for ZIP file.

        Args:
            filename: Path to ZIP file

        Returns:
            ZipFile object
        """
        if filename in self.handle:
            return self.handle.get(filename, None)
        else:
            zfile = zipfile.ZipFile(filename)
            self.handle[filename] = zfile
            return zfile

    def __get_handle_7z(self, filename, password=None):
        """Get or create handle for 7Z file.

        Args:
            filename: Path to 7Z file
            password: Optional password for encrypted archives

        Returns:
            SevenZipFile object
        """
        if filename in self.handle:
            return self.handle.get(filename, None)
        else:
            if password:
                zfile = py7zr.SevenZipFile(filename, mode="r", password=password)
            else:
                zfile = py7zr.SevenZipFile(filename, mode="r")
            self.handle[filename] = zfile
            return zfile

    def format(self, filehandle, filename, filename_ex):
        """Analyze and detect archive format.

        Args:
            filehandle: File data (memory mapped)
            filename: Path to archive file
            filename_ex: Extended filename info

        Returns:
            Dictionary with format info, or None if not recognized
        """
        ret = {}

        try:
            mm = filehandle
            if mm[:4] == b"PK\x03\x04":  # ZIP signature
                with contextlib.suppress(zipfile.BadZipfile):
                    with zipfile.ZipFile(filename) as zfile:
                        names = zfile.namelist()

                        if names is not None:
                            for name in names:
                                n = name.lower()
                                if n == "classes.dex":
                                    ret["ff_apk"] = "apk"
                                elif n == "ppt/presentation.xml":
                                    ret["ff_ooxml"] = "pptx"
                                elif n == "word/document.xml":
                                    ret["ff_ooxml"] = "docx"
                                elif n == "xl/workbook.xml":
                                    ret["ff_ooxml"] = "xlsx"
                                elif n == "mimetype":
                                    data = zfile.read(name)
                                    if data == b"application/hwp+zip":
                                        ret["ff_hwpx"] = "hwpx"

                            ret["ff_zip"] = "zip"
                return ret

            elif mm[:4] == b"7z\xbc\xaf":  # 7Z signature
                ret["ff_7z"] = "7z"
                return ret

        except (IOError, OSError) as e:
            logger.debug("Format detection IO error for %s: %s", filename, e)
        except Exception as e:
            logger.warning("Unexpected error in format detection for %s: %s", filename, e)

        return None

    def arclist(self, filename, fileformat, password=None):
        """List files in the archive.

        Args:
            filename: Path to archive file
            fileformat: Format info from format() method
            password: Optional password for encrypted archives

        Returns:
            List of [engine_id, filename] pairs
        """
        file_scan_list = []

        # Store password for later use in unarc
        self.password = password

        try:
            if "ff_zip" in fileformat:
                zfile = self.__get_handle(filename)
                for name in zfile.namelist():
                    # CWE-22: Path traversal prevention
                    if k2security.is_safe_archive_member(name):
                        file_scan_list.append(["arc_zip", name])

            elif "ff_7z" in fileformat:
                # 7z needs password at open time for encrypted headers
                zfile = self.__get_handle_7z(filename, password)
                for name in zfile.getnames():
                    # CWE-22: Path traversal prevention
                    if k2security.is_safe_archive_member(name):
                        file_scan_list.append(["arc_7z", name])

        except (IOError, OSError) as e:
            logger.debug("Archive list IO error for %s: %s", filename, e)
        except Exception as e:
            logger.warning("Unexpected error listing archive %s: %s", filename, e)

        return file_scan_list

    def unarc(self, arc_engine_id, arc_name, fname_in_arc):
        """Extract a file from the archive.

        Args:
            arc_engine_id: Engine ID ('arc_zip' or 'arc_7z')
            arc_name: Path to archive file
            fname_in_arc: Name of file to extract

        Returns:
            Extracted file data, or None on error
        """
        # CWE-22: Path traversal prevention
        if not k2security.is_safe_archive_member(fname_in_arc):
            logger.warning("Unsafe archive member rejected: %s in %s", fname_in_arc, arc_name)
            return None

        if arc_engine_id == "arc_zip":
            return self.__unarc_zip(arc_name, fname_in_arc)
        elif arc_engine_id == "arc_7z":
            return self.__unarc_7z(arc_name, fname_in_arc)

        return None

    def __unarc_zip(self, arc_name, fname_in_arc):
        """Extract a file from ZIP archive with password support.

        Args:
            arc_name: Path to archive file
            fname_in_arc: Name of file to extract

        Returns:
            Extracted file data, or None on error
        """
        zfile = self.__get_handle(arc_name)
        if zfile is None:
            return None

        # Try 1: Extract without password first
        try:
            return zfile.read(fname_in_arc)
        except RuntimeError as e:
            # RuntimeError is raised for password-protected files
            if "password" not in str(e).lower():
                logger.debug("ZIP extract error for %s: %s", fname_in_arc, e)
                return None
        except zipfile.BadZipfile as e:
            logger.debug("Bad ZIP file when extracting %s: %s", fname_in_arc, e)
            return None
        except (IOError, OSError) as e:
            logger.debug("Archive extract IO error for %s in %s: %s", fname_in_arc, arc_name, e)
            return None
        except Exception as e:
            logger.debug("Archive extract error for %s in %s: %s", fname_in_arc, arc_name, e)
            return None

        # Try 2: Extract with stored password
        if self.password:
            try:
                pwd = self.password.encode() if isinstance(self.password, str) else self.password
                return zfile.read(fname_in_arc, pwd=pwd)
            except RuntimeError as e:
                logger.debug("Wrong password for %s in ZIP archive: %s", fname_in_arc, e)
                raise RuntimeError("password required")
            except zipfile.BadZipfile as e:
                logger.debug("Bad ZIP file when extracting %s with password: %s", fname_in_arc, e)
            except (IOError, OSError) as e:
                logger.debug("Archive extract IO error for %s in %s: %s", fname_in_arc, arc_name, e)
            except Exception as e:
                logger.debug("Archive extract error for %s in %s: %s", fname_in_arc, arc_name, e)
        else:
            logger.debug("Password required for %s in ZIP archive", fname_in_arc)
            raise RuntimeError("password required")

        return None

    def __unarc_7z(self, arc_name, fname_in_arc):
        """Extract a file from 7Z archive with password support.

        Args:
            arc_name: Path to archive file
            fname_in_arc: Name of file to extract

        Returns:
            Extracted file data, or None on error
        """
        # 7z handle is already created with password in arclist
        zfile = self.handle.get(arc_name)
        if zfile is None:
            return None

        try:
            zfile.reset()  # Reset required because EOF may be exhausted
            with tempfile.TemporaryDirectory() as tmpdir:
                zfile.extract(path=tmpdir, targets=[fname_in_arc])
                extracted_path = os.path.join(tmpdir, fname_in_arc)
                if os.path.exists(extracted_path):
                    with open(extracted_path, "rb") as f:
                        return f.read()
        except py7zr.exceptions.PasswordRequired:
            logger.debug("Password required for %s in 7Z archive", fname_in_arc)
            raise RuntimeError("password required")
        except (IOError, OSError) as e:
            logger.debug("Archive extract IO error for %s in %s: %s", fname_in_arc, arc_name, e)
        except Exception as e:
            logger.debug("Archive extract error for %s in %s: %s", fname_in_arc, arc_name, e)

        return None

    def arcclose(self):
        """Close all open archive handles."""
        for fname in list(self.handle.keys()):
            try:
                zfile = self.handle[fname]
                if hasattr(zfile, "close"):
                    zfile.close()
            except (IOError, OSError) as e:
                logger.debug("Archive close IO error for %s: %s", fname, e)
            except Exception as e:
                logger.debug("Archive close error for %s: %s", fname, e)
            finally:
                self.handle.pop(fname, None)

    def mkarc(self, arc_engine_id, arc_name, file_infos):
        """Create or update an archive.

        Args:
            arc_engine_id: Engine ID ('arc_zip' or 'arc_7z')
            arc_name: Path to archive file
            file_infos: List of file info structures

        Returns:
            True if successful, False otherwise
        """
        try:
            if arc_engine_id == "arc_zip":
                with zipfile.ZipFile(arc_name, "w") as zfile:
                    for file_info in file_infos:
                        rname = file_info.get_filename()
                        with contextlib.suppress(IOError):
                            with open(rname, "rb") as fp:
                                buf = fp.read()
                                a_name = file_info.get_filename_in_archive()
                                zfile.writestr(a_name, buf, compress_type=zipfile.ZIP_DEFLATED)
                return True

            elif arc_engine_id == "arc_7z":
                return self._convert_7z_to_zip(arc_name, file_infos)

        except (IOError, OSError) as e:
            logger.error("Archive creation IO error for %s: %s", arc_name, e)
        except Exception as e:
            logger.error("Unexpected error creating archive %s: %s", arc_name, e)

        return False

    def _convert_7z_to_zip(self, arc_name, file_infos):
        """Convert 7Z archive to ZIP format and update with cured files.

        Args:
            arc_name: Archive file name
            file_infos: Information about files to be included

        Returns:
            True if successful, False otherwise
        """
        uname = kavutil.uniq_string()
        tname = f".tmp_{uname}"
        oname = f".org_{uname}"

        try:
            with open(arc_name, "rb") as f:
                if f.read(2) != b"PK":
                    self._convert_7z_file_format(arc_name, tname, oname)

            shutil.move(arc_name, arc_name + oname)
            shutil.move(arc_name + tname, arc_name)

            # Put the cured file into the converted ZIP file
            kavutil.make_zip(arc_name, file_infos)

            # Clean up original file
            fname = arc_name + oname
            if os.path.exists(fname):
                with contextlib.suppress(k2security.SecurityError):
                    fname_dir = os.path.dirname(fname) or os.getcwd()
                    k2security.safe_remove_file(fname, fname_dir)

            return True

        except (IOError, OSError) as e:
            logger.error("7Z to ZIP conversion IO error for %s: %s", arc_name, e)
        except Exception as e:
            logger.error("Unexpected error converting 7Z to ZIP %s: %s", arc_name, e)

        return False

    def _convert_7z_file_format(self, arc_name, tname, oname):
        """Convert 7Z file to ZIP format while preserving contents.

        Args:
            arc_name: Original 7Z archive name
            tname: Temporary file suffix
            oname: Original file backup suffix
        """
        with zipfile.ZipFile(arc_name + tname, "w") as zfile_w:
            with py7zr.SevenZipFile(arc_name, mode="r") as zfile_r:
                with tempfile.TemporaryDirectory() as tmpdir:
                    zfile_r.extractall(path=tmpdir)
                    for name in zfile_r.getnames():
                        extracted_path = os.path.join(tmpdir, name)
                        if os.path.isfile(extracted_path):
                            with open(extracted_path, "rb") as f:
                                data = f.read()
                            zfile_w.writestr(name, data, compress_type=zipfile.ZIP_DEFLATED)



# -*- coding:utf-8 -*-
# Author: Kei Choi(hanul93@gmail.com)

"""
HTML File Format Engine Plugin

This plugin handles HTML format for scanning, malware detection, and extraction.
"""

import contextlib
import logging
import os
import re

from kicomav.plugins import kavutil
from kicomav.plugins import kernel
from kicomav.kavcore import k2security
from kicomav.kavcore.plugin_base import ArchivePluginBase

# Module logger
logger = logging.getLogger(__name__)

HTML_KEY_COUNT = 3  # Is there at least 3 HTML keywords?


# -------------------------------------------------------------------------
# KavMain class
# -------------------------------------------------------------------------
class KavMain(ArchivePluginBase):
    """HTML malware detector and archive handler plugin.

    This plugin provides functionality for:
    - Detecting HTML format
    - Scanning for HTML-based malware
    - Extracting scripts from HTML files
    """

    def __init__(self):
        """Initialize the HTML plugin."""
        super().__init__(
            author="Kei Choi",
            version="1.0",
            title="HTML Engine",
            kmd_name="html",
        )
        self.p_html = None
        self.p_script = None
        self.p_html_malware = None

    def _custom_init(self) -> int:
        """Custom initialization for HTML plugin.

        Returns:
            0 for success
        """
        pat = r"<\s*html\b|\bdoctype\b|<\s*head\b|<\s*title\b|<\s*meta\b|\bhref\b|<\s*link\b|<\s*body\b|<\s*script\b|<\s*iframe\b"
        self.p_html = re.compile(pat, re.IGNORECASE)

        # script, iframe, php keywords
        pat = rb"<script.*?>[\d\D]*?</script>|<iframe.*?>[\d\D]*?</iframe>|<\?(php\b)?[\d\D]*?\?>"
        self.p_script = re.compile(pat, re.IGNORECASE)

        # HTML malware pattern
        self.p_html_malware = re.compile(rb"\?ob_start.+?>\s*<iframe")

        return 0

    def getinfo(self):
        """Get plugin information.

        Returns:
            Dictionary containing plugin metadata
        """
        info = super().getinfo()
        info["sig_num"] = 1
        return info

    def listvirus(self):
        """Get list of detectable viruses.

        Returns:
            List of virus names
        """
        return ["Trojan.HTML.IFrame.a"]

    def format(self, filehandle, filename, filename_ex):
        """Analyze and detect HTML format.

        Args:
            filehandle: File data (memory mapped)
            filename: Path to file
            filename_ex: Extended filename info

        Returns:
            Dictionary with format info, or None if not recognized
        """
        try:
            if filename_ex:
                with contextlib.suppress(IndexError):
                    if filename_ex.split("/")[-2] == "HTML":
                        return None

            mm = filehandle
            buf = mm[:4096]

            if kavutil.is_textfile(buf):
                buf = buf.decode("latin-1")
                ret = self.p_html.findall(buf) if self.p_html else []
                if len(set(ret)) >= HTML_KEY_COUNT:
                    fileformat = {"keyword": list(set(ret))}
                    return {"ff_html": fileformat}

        except (IOError, OSError) as e:
            logger.debug("Format detection IO error for %s: %s", filename, e)
        except Exception as e:
            logger.warning("Unexpected error in format detection for %s: %s", filename, e)

        return None

    def scan(self, filehandle, filename, fileformat, filename_ex):
        """Scan for malware.

        Args:
            filehandle: File data (memory mapped)
            filename: Path to file
            fileformat: Format info from format() method
            filename_ex: Extended filename info

        Returns:
            Tuple of (found, malware_name, malware_id, result)
        """
        try:
            mm = filehandle
            buf = mm.read(4096) if hasattr(mm, "read") else mm[:4096]

            if kavutil.is_textfile(buf) and self.p_html_malware and self.p_html_malware.search(buf):
                return True, "Trojan.HTML.IFrame.a", 0, kernel.INFECTED

        except (IOError, OSError) as e:
            logger.debug("Scan IO error for %s: %s", filename, e)
        except Exception as e:
            logger.warning("Unexpected error scanning %s: %s", filename, e)

        return False, "", -1, kernel.NOT_FOUND

    def disinfect(self, filename, malware_id):
        """Disinfect malware.

        Args:
            filename: Path to infected file
            malware_id: Malware ID to disinfect

        Returns:
            True if successful, False otherwise
        """
        try:
            if malware_id == 0:
                filename_dir = os.path.dirname(filename) or os.getcwd()
                k2security.safe_remove_file(filename, filename_dir)
                return True

        except (IOError, OSError, k2security.SecurityError) as e:
            logger.debug("Disinfect error for %s: %s", filename, e)
        except Exception as e:
            logger.warning("Unexpected error disinfecting %s: %s", filename, e)

        return False

    def arclist(self, filename, fileformat, password=None):
        """List files in the archive (scripts in HTML).

        Args:
            filename: Path to HTML file
            fileformat: Format info from format() method

        Returns:
            List of [engine_id, filename] pairs
        """
        file_scan_list = []

        if "ff_html" not in fileformat:
            return file_scan_list

        try:
            with open(filename, "rb") as fp:
                buf = fp.read()

            s_count = 1
            i_count = 1
            p_count = 1

            if self.p_script:
                for obj in self.p_script.finditer(buf):
                    t = obj.group()
                    p = t.lower()

                    if p.find(b"<script") != -1:
                        file_scan_list.append(["arc_html", "HTML/Script #%d" % s_count])
                        s_count += 1
                    elif p.find(b"<iframe") != -1:
                        file_scan_list.append(["arc_html", "HTML/IFrame #%d" % i_count])
                        i_count += 1
                    elif p.find(b"<?") != -1:
                        file_scan_list.append(["arc_html", "HTML/PHP #%d" % p_count])
                        p_count += 1

        except (IOError, OSError) as e:
            logger.debug("Archive list IO error for %s: %s", filename, e)
        except Exception as e:
            logger.warning("Unexpected error listing archive %s: %s", filename, e)

        return file_scan_list

    def unarc(self, arc_engine_id, arc_name, fname_in_arc):
        """Extract a file from the archive (script from HTML).

        Args:
            arc_engine_id: Engine ID ('arc_html')
            arc_name: Path to HTML file
            fname_in_arc: Name of script to extract

        Returns:
            Extracted script data, or None on error
        """
        if arc_engine_id != "arc_html":
            return None

        try:
            with open(arc_name, "rb") as fp:
                buf = fp.read()

            s_count = 1
            i_count = 1
            p_count = 1

            if self.p_script:
                for obj in self.p_script.finditer(buf):
                    t = obj.group()
                    pos = obj.span()
                    p = t.lower()

                    if p.find(b"<script") != -1:
                        k = "HTML/Script #%d" % s_count
                        s_count += 1
                    elif p.find(b"<iframe") != -1:
                        k = "HTML/IFrame #%d" % i_count
                        i_count += 1
                    elif p.find(b"<?") != -1:
                        k = "HTML/PHP #%d" % p_count
                        p_count += 1
                    else:
                        k = ""

                    if k == fname_in_arc:
                        return buf[pos[0] : pos[1]]

        except (IOError, OSError) as e:
            logger.debug("Archive extract error for %s in %s: %s", fname_in_arc, arc_name, e)
        except Exception as e:
            logger.warning("Unexpected error extracting %s from %s: %s", fname_in_arc, arc_name, e)

        return None

    def mkarc(self, arc_engine_id, arc_name, file_infos):
        """Create an archive.

        Args:
            arc_engine_id: Engine ID ('arc_html')
            arc_name: Path to archive file
            file_infos: List of file info structures

        Returns:
            True if successful, False otherwise
        """
        if arc_engine_id != "arc_html":
            return False

        try:
            all_script_info = []

            with open(arc_name, "rb") as fp:
                buf = fp.read()

            if self.p_script:
                for obj in self.p_script.finditer(buf):
                    t = obj.group()
                    pos = obj.span()
                    p = t.lower()

                    if p.find(b"<script") != -1:
                        all_script_info.append(["script", pos, t])
                    elif p.find(b"<iframe") != -1:
                        all_script_info.append(["iframe", pos, t])
                    elif p.find(b"<?") != -1:
                        all_script_info.append(["php", pos, t])

            org_buf = buf

            for idx, file_info in enumerate(file_infos):
                rname = file_info.get_filename()
                with contextlib.suppress(IOError):
                    if os.path.exists(rname):
                        with open(rname, "rb") as fp:
                            new_buf = fp.read()

                            if len(all_script_info[idx][2]) < len(new_buf):
                                return False

                            new_buf += b" " * (len(all_script_info[idx][2]) - len(new_buf))
                            all_script_info[idx][2] = new_buf
                    else:
                        new_buf = b" " * len(all_script_info[idx][2])
                        all_script_info[idx][2] = new_buf

            with open(arc_name, "wb") as fp:
                start_pos = 0
                for script_info in all_script_info:
                    pos = script_info[1]
                    fp.write(org_buf[start_pos : pos[0]])
                    fp.write(script_info[2])
                    start_pos = pos[1]

                fp.write(org_buf[start_pos:])

            return True

        except (IOError, OSError) as e:
            logger.error("Archive creation IO error for %s: %s", arc_name, e)
        except Exception as e:
            logger.error("Unexpected error creating archive %s: %s", arc_name, e)

        return False

    def arcclose(self):
        """Close all open archive handles."""
        pass  # No persistent handles to close



# -*- coding:utf-8 -*-
# Author: Kei Choi(hanul93@gmail.com)

"""
KicomAV Updater Module

This module provides update functionality for KicomAV signatures and package.
"""

import contextlib
import datetime
import glob
import gzip
import hashlib
import logging
import os
import re
from dataclasses import dataclass, field
from pathlib import Path
from typing import Callable, List, Optional, Set, Tuple

from . import k2security
from .config import get_config

# Module logger
logger = logging.getLogger(__name__)

# Default update URL
DEFAULT_UPDATE_URL = "https://raw.githubusercontent.com/hanul93/kicomav-db/master/update/"


@dataclass
class UpdateResult:
    """Result of an update operation.

    Attributes:
        success: Whether the update completed successfully
        updated_files: List of files that were updated
        removed_files: List of orphan files that were removed
        errors: List of error messages encountered
        package_update_available: Whether a newer kicomav package is available
        latest_version: Latest available package version (if checked)
        current_version: Currently installed package version
    """

    success: bool = True
    updated_files: List[str] = field(default_factory=list)
    removed_files: List[str] = field(default_factory=list)
    errors: List[str] = field(default_factory=list)
    package_update_available: bool = False
    latest_version: Optional[str] = None
    current_version: Optional[str] = None


# Type alias for progress callback
ProgressCallback = Callable[[str, str], None]


def get_installed_version(package_name: str = "kicomav") -> Optional[str]:
    """Get the installed version of a package.

    Args:
        package_name: Name of the package to check

    Returns:
        Version string or None if not installed
    """
    try:
        from importlib.metadata import PackageNotFoundError, version

        return version(package_name)
    except PackageNotFoundError:
        return None
    except Exception as e:
        logger.debug("Failed to get installed version: %s", e)
        return None


def get_latest_version(package_name: str = "kicomav") -> Optional[str]:
    """Get the latest version of a package from PyPI.

    Args:
        package_name: Name of the package to check

    Returns:
        Latest version string or None if unavailable
    """
    try:
        import requests

        response = requests.get(f"https://pypi.org/pypi/{package_name}/json", timeout=10)
        if response.status_code == 404:
            return None
        response.raise_for_status()
        return response.json()["info"]["version"]
    except Exception as e:
        logger.debug("Failed to get latest version from PyPI: %s", e)
        return None


def check_package_update(
    package_name: str = "kicomav",
) -> Tuple[bool, Optional[str], Optional[str]]:
    """Check if a package update is available.

    Args:
        package_name: Name of the package to check

    Returns:
        Tuple of (update_available, current_version, latest_version)
    """
    current = get_installed_version(package_name)
    latest = get_latest_version(package_name)

    if latest is None:
        return False, current, None

    if current is None:
        return True, None, latest

    update_available = current != latest
    return update_available, current, latest


def _check_need_update(filepath: str, expected_hash: str) -> bool:
    """Check if a file needs to be updated by comparing SHA1 hashes.

    Args:
        filepath: Path to the local file
        expected_hash: Expected SHA1 hash from update config

    Returns:
        True if update is needed, False otherwise
    """
    try:
        with open(filepath, "rb") as fp:
            data = fp.read()
        s = hashlib.sha1()
        s.update(data)
        return s.hexdigest().lower() != expected_hash.lower()
    except (IOError, OSError):
        return True  # File doesn't exist or can't be read


def _download_file(
    url: str,
    filename: str,
    path: str,
    gz: bool = False,
    progress_callback: Optional[ProgressCallback] = None,
) -> bool:
    """Download a single file.

    Args:
        url: Base URL for download
        filename: Name of file to download (can include subdirectory path)
        path: Local directory to save the file
        gz: Whether the file is gzipped on the server
        progress_callback: Optional callback for progress updates

    Returns:
        True if download succeeded, False otherwise
    """
    # Normalize path separators
    filename = filename.replace("\\", "/")

    # Split into directory path and actual filename
    if "/" in filename:
        subdir, actual_filename = filename.rsplit("/", 1)
        # Validate for path traversal
        if ".." in subdir:
            logger.warning("Path traversal detected in filename: %s", filename)
            return False
    else:
        subdir = ""
        actual_filename = filename

    # CWE-434: Validate filename before download
    try:
        k2security.validate_download_filename(actual_filename)
    except k2security.SecurityError as e:
        logger.warning("Invalid filename %s: %s", filename, e)
        return False

    # Create subdirectory if needed
    if subdir:
        subdir_path = os.path.join(path, subdir)
        os.makedirs(subdir_path, exist_ok=True)

    # Build download URL
    download_url = url + filename.replace("\\", "/")
    if gz:
        download_url += ".gz"

    # Get full path for saving
    save_path = os.path.join(path, filename)
    if gz:
        save_path += ".gz"

    if progress_callback:
        progress_callback(filename, "downloading")

    # Download the file
    try:
        base_path = os.path.join(path, subdir) if subdir else path
        k2security.safe_download_file(download_url, save_path, base_path)
    except k2security.SecurityError as e:
        logger.warning("Download failed for %s: %s", filename, e)
        return False

    # Handle gzipped files
    if gz:
        try:
            with gzip.open(save_path, "rb") as gz_file:
                data = gz_file.read()
            final_path = os.path.join(path, filename)
            with open(final_path, "wb") as out_file:
                out_file.write(data)
            # Delete the gz file
            base_path = os.path.join(path, subdir) if subdir else path
            with contextlib.suppress(k2security.SecurityError):
                k2security.safe_remove_file(save_path, base_path)
        except Exception as e:
            logger.warning("Failed to decompress %s: %s", filename, e)
            return False

    if progress_callback:
        progress_callback(filename, "updated")

    return True


def _get_signature_download_list(url: str, rules_path: str) -> Tuple[List[str], Set[str]]:
    """Get list of signature files that need updating.

    Args:
        url: Base URL for update config
        rules_path: Local rules directory path

    Returns:
        Tuple of (files_to_download, all_files_in_config)
    """
    down_list = []
    cfg_files: Set[str] = set()

    if not rules_path:
        return down_list, cfg_files

    try:
        # Download update.cfg to temp location
        temp_cfg_path = os.path.join(rules_path, "update.cfg")
        _download_file(url, "update.cfg", rules_path)

        with open(temp_cfg_path, "r") as f:
            buf = f.read()

        # Format: [sha1] [filepath]
        p_lists = re.compile(r"([A-Fa-f0-9]{40}) (.+)")
        lines = p_lists.findall(buf)

        for line in lines:
            fhash = line[0]
            fname = line[1].strip()

            cfg_files.add(fname)

            # Normalize path separators
            fname_normalized = fname.replace("/", os.sep).replace("\\", os.sep)
            local_file = os.path.join(rules_path, fname_normalized)

            # Special handling for whitelist.txt
            if fname == "whitelist.txt":
                if os.path.exists(local_file):
                    continue  # Keep local whitelist
                else:
                    down_list.append(fname)
                    continue

            # Check if update needed
            if _check_need_update(local_file, fhash):
                down_list.append(fname)

        # Keep update.cfg for timestamp tracking (do not delete)

    except Exception as e:
        logger.debug("Failed to get download list: %s", e)

    return down_list, cfg_files


def _get_local_files(rules_path: str) -> Set[str]:
    """Get all local files in rules path.

    Args:
        rules_path: Local rules directory path

    Returns:
        Set of relative file paths
    """
    local_files: Set[str] = set()
    for filepath in glob.glob(os.path.join(rules_path, "**", "*"), recursive=True):
        if os.path.isdir(filepath):
            continue
        rel_path = os.path.relpath(filepath, rules_path)
        rel_path = rel_path.replace("\\", "/")
        local_files.add(rel_path)
    return local_files


def _remove_orphan_files(
    rules_path: str,
    cfg_files: Set[str],
    progress_callback: Optional[ProgressCallback] = None,
) -> List[str]:
    """Remove local files not in update.cfg.

    Args:
        rules_path: Local rules directory path
        cfg_files: Set of files listed in update.cfg
        progress_callback: Optional progress callback

    Returns:
        List of removed files
    """
    removed = []
    local_files = _get_local_files(rules_path)
    cfg_files_normalized = {f.replace("\\", "/") for f in cfg_files}
    orphan_files = local_files - cfg_files_normalized

    for orphan in orphan_files:
        if orphan in ("whitelist.txt", "update.cfg"):
            continue

        orphan_path = os.path.join(rules_path, orphan)
        try:
            if "/" in orphan:
                subdir = orphan.rsplit("/", 1)[0]
                base_path = os.path.join(rules_path, subdir)
            else:
                base_path = rules_path

            k2security.safe_remove_file(orphan_path, base_path)
            removed.append(orphan)
            if progress_callback:
                progress_callback(orphan, "removed")
        except k2security.SecurityError:
            pass

    return removed


def _remove_empty_dirs(rules_path: str) -> None:
    """Remove empty directories in rules path.

    Args:
        rules_path: Local rules directory path
    """
    for root, dirs, files in os.walk(rules_path, topdown=False):
        for dir_name in dirs:
            dir_path = os.path.join(root, dir_name)
            try:
                if not os.listdir(dir_path):
                    os.rmdir(dir_path)
            except OSError:
                pass


def update_signatures(
    rules_path: Optional[str] = None,
    url: Optional[str] = None,
    progress_callback: Optional[ProgressCallback] = None,
) -> UpdateResult:
    """Update signature files.

    Args:
        rules_path: Path to rules directory. If None, uses SYSTEM_RULES_BASE from config.
        url: Update server URL. If None, uses default URL.
        progress_callback: Optional callback called with (filename, status) for each file.

    Returns:
        UpdateResult with details of the update operation
    """
    result = UpdateResult()

    # Get rules path from config if not provided
    if rules_path is None:
        config = get_config()
        rules_path = config.system_rules_base

    if not rules_path:
        result.success = False
        result.errors.append("No rules path configured (SYSTEM_RULES_BASE not set)")
        return result

    if not os.path.exists(rules_path):
        result.success = False
        result.errors.append(f"Rules path does not exist: {rules_path}")
        return result

    # Use default URL if not provided
    if url is None:
        url = DEFAULT_UPDATE_URL

    try:
        # Get list of files to download
        down_list, cfg_files = _get_signature_download_list(url, rules_path)

        # Remove orphan files
        if cfg_files:
            removed = _remove_orphan_files(rules_path, cfg_files, progress_callback)
            result.removed_files = removed
            _remove_empty_dirs(rules_path)

        # Download updated files
        for filename in down_list:
            if _download_file(url, filename, rules_path, gz=False, progress_callback=progress_callback):
                result.updated_files.append(filename)
            else:
                result.errors.append(f"Failed to download: {filename}")

    except Exception as e:
        result.success = False
        result.errors.append(f"Update failed: {str(e)}")
        logger.exception("Signature update failed")

    return result


def update(progress_callback: Optional[ProgressCallback] = None, check_package: bool = True) -> UpdateResult:
    """Full update: check package version and update signatures.

    Args:
        progress_callback: Optional callback for progress updates
        check_package: Whether to check for package updates

    Returns:
        UpdateResult with details of the update operation
    """
    result = UpdateResult()

    # Check package update
    if check_package:
        update_available, current, latest = check_package_update()
        result.package_update_available = update_available
        result.latest_version = latest
        result.current_version = current

    # Update signatures
    sig_result = update_signatures(progress_callback=progress_callback)

    # Merge signature update results
    result.success = sig_result.success
    result.updated_files = sig_result.updated_files
    result.removed_files = sig_result.removed_files
    result.errors.extend(sig_result.errors)

    return result


def get_update_timestamp(rules_path: Optional[str] = None) -> Optional[datetime.datetime]:
    """Read UTC timestamp from update.cfg first line.

    The update.cfg file is expected to have a timestamp comment on the first line:
    # Sat Dec 27 03:05:44 2025 UTC

    Args:
        rules_path: Path to rules directory. If None, uses system_rules_base from config.

    Returns:
        datetime object with UTC timezone or None if not found or parse failed
    """
    if rules_path is None:
        config = get_config()
        rules_path = config.system_rules_base

    if not rules_path:
        return None

    cfg_path = os.path.join(rules_path, "update.cfg")
    if not os.path.exists(cfg_path):
        return None

    try:
        with open(cfg_path, "r", encoding="utf-8") as f:
            first_line = f.readline().strip()

        # Parse format: # Sat Dec 27 03:05:44 2025 UTC
        if first_line.startswith("#"):
            time_str = first_line[1:].strip()  # Remove "#"
            if time_str.endswith(" UTC"):
                time_str = time_str[:-4]  # Remove " UTC"
                return datetime.datetime.strptime(time_str, "%a %b %d %H:%M:%S %Y").replace(
                    tzinfo=datetime.timezone.utc
                )
    except Exception as e:
        logger.debug("Failed to parse update.cfg timestamp: %s", e)

    return None


def get_last_update_time() -> datetime.datetime:
    """Get last update time from update.cfg or fallback to __last_update__.

    Returns:
        datetime object with UTC timezone
    """
    cfg_time = get_update_timestamp()
    if cfg_time:
        return cfg_time

    # Fallback to __last_update__ from kicomav package
    try:
        from kicomav import __last_update__

        time_str = __last_update__
        if time_str.endswith(" UTC"):
            time_str = time_str[:-4]
        return datetime.datetime.strptime(time_str, "%a %b %d %H:%M:%S %Y").replace(tzinfo=datetime.timezone.utc)
    except Exception as e:
        logger.debug("Failed to parse __last_update__: %s", e)
        # Ultimate fallback: epoch time
        return datetime.datetime(1970, 1, 1, tzinfo=datetime.timezone.utc)



# -*- coding:utf-8 -*-
# Author: Kei Choi(hanul93@gmail.com)

"""
PDF File Format Engine Plugin

This plugin handles PDF format for scanning, malware detection, and extraction.
"""

import contextlib
import logging
import os
import re
from pathlib import Path

from kicomav.plugins import kernel
from kicomav.kavcore import k2security
from kicomav.kavcore.plugin_base import ArchivePluginBase

# Module logger
logger = logging.getLogger(__name__)


# -------------------------------------------------------------------------
# PdfFile class
# -------------------------------------------------------------------------
class PdfFile:
    def __init__(self, filename, verbose=False):
        self.REF_OBJ = 0  # Extract Stream from referenced OBJ
        self.IN_OBJ = 1  # Extract Stream from OBJ containing the sentence

        # Try to scan for malware if the pattern exists
        self.p_pdf_scanables = {}
        pats = {rb"/JS\s+(\d+)\s+0\s+R\b": self.REF_OBJ, rb"/Length\s+0\b": self.IN_OBJ}

        for pat in pats:
            self.p_pdf_scanables[re.compile(pat, re.IGNORECASE)] = pats[pat]

        # Use to record the location of OBJ within PDF
        s = rb"(\d+)\s+0\s+obj\s*<<.+?endobj"
        self.p_obj = re.compile(s, re.IGNORECASE | re.DOTALL)
        self.pdf_obj_off = None

        # Extract Stream
        pat = rb"stream\s*(.+?)\s*endstream"
        self.p_stream = re.compile(pat, re.IGNORECASE | re.DOTALL)

        # /Filter
        pat = rb"/Filter\s*/(\w+)"
        self.p_pdf_filter = re.compile(pat, re.IGNORECASE)

        self.buf = None
        self.verbose = verbose  # Debugging mode
        self.filename = filename

        self.p = re.compile(rb"[A-Fa-f0-9]+")

        self.num_obj = 0  # Number of obj inserted into PDF
        self.objdata = {}  # objects data
        self.pdf_obj_off = {}
        self.payloads = []  # Extracted payloads
        self.parse()

    def parse(self):
        self.payloads = self.extract_objects_data()
        for _, no in enumerate(self.payloads):
            fname = f"Obj #{int(no)}"
            self.objdata[fname] = no

        self.num_obj = len(self.payloads)

    def extract_objects_data(self):
        payloads: list[(bytes, bytes)] = []

        pdf_bytes = Path(self.filename).read_bytes()
        self.buf = pdf_bytes

        for pat in self.p_pdf_scanables.keys():
            for p in pat.finditer(pdf_bytes):
                if self.p_pdf_scanables[pat] == self.REF_OBJ:
                    payloads.append(p.groups()[0])
                else:  # self.IN_OBJ
                    self.__search_object_off(pdf_bytes)

                    for obj_no in self.pdf_obj_off.keys():
                        start_off = self.pdf_obj_off[obj_no][0]
                        end_off = self.pdf_obj_off[obj_no][1]
                        if start_off < p.span()[0] < end_off:
                            payloads.append(obj_no)
                            break

        return payloads

    def close(self):
        pass

    def namelist(self):
        names = []

        if self.num_obj:  # Is there a target for extracting Stream?
            names = list(self.objdata.keys())
            names.sort(key=lambda x: int(x.split("#")[1]))

        return names

    def read(self, fname):
        obj_no = self.objdata.get(fname, None)
        if obj_no is None:
            return None

        start_off, end_off = self.pdf_obj_off[obj_no]

        p = self.p_stream.search(self.buf[start_off:end_off])
        if p is None:
            return None

        return p.groups()[0]

    def __search_object_off(self, buf):
        """Record the location of PDF OBJ."""
        if self.pdf_obj_off:
            return

        self.pdf_obj_off = {}
        for p in self.p_obj.finditer(buf):
            obj_no = p.groups()[0]
            obj_off = p.span()
            self.pdf_obj_off[obj_no] = obj_off

        if not self.pdf_obj_off:
            self.pdf_obj_off = None


# -------------------------------------------------------------------------
# KavMain class
# -------------------------------------------------------------------------
class KavMain(ArchivePluginBase):
    """PDF malware detector and archive handler plugin.

    This plugin provides functionality for:
    - Detecting PDF format
    - Scanning for PDF-based malware
    - Extracting objects from PDF files
    """

    def __init__(self):
        """Initialize the PDF plugin."""
        super().__init__(
            author="Kei Choi",
            version="1.2",
            title="PDF Engine",
            kmd_name="pdf",
        )
        self.p_pdf_header = None
        self.p_pdf_trojan_js = None

    def _custom_init(self) -> int:
        """Custom initialization for PDF plugin.

        Returns:
            0 for success
        """
        # PDF header
        pat = rb"^s*%PDF-1."
        self.p_pdf_header = re.compile(pat, re.IGNORECASE)

        # PDF Trojan pattern for detection
        pat = rb"this\.exportDataObject.+?cName:.+?nLaunch"
        self.p_pdf_trojan_js = re.compile(pat)

        return 0

    def getinfo(self):
        """Get plugin information.

        Returns:
            Dictionary containing plugin metadata
        """
        info = super().getinfo()
        info["sig_num"] = 1
        return info

    def listvirus(self):
        """Get list of detectable viruses.

        Returns:
            List of virus names
        """
        return ["Trojan.PDF.Generic"]

    def __get_handle(self, filename):
        """Get or create handle for PDF file.

        Args:
            filename: Path to PDF file

        Returns:
            PdfFile object or None
        """
        if filename in self.handle:
            return self.handle.get(filename, None)

        try:
            zfile = PdfFile(filename, self.verbose)
            self.handle[filename] = zfile
            return zfile

        except (IOError, OSError) as e:
            logger.debug("Failed to open PDF file %s: %s", filename, e)
        except Exception as e:
            logger.warning("Unexpected error opening PDF file %s: %s", filename, e)

        return None

    def format(self, filehandle, filename, filename_ex):
        """Analyze and detect PDF format.

        Args:
            filehandle: File data (memory mapped)
            filename: Path to file
            filename_ex: Extended filename info

        Returns:
            Dictionary with format info, or None if not recognized
        """
        try:
            mm = filehandle
            buf = mm[:4096]

            if self.p_pdf_header and self.p_pdf_header.match(buf):
                return {"ff_pdf": "PDF"}

        except (IOError, OSError) as e:
            logger.debug("Format detection IO error for %s: %s", filename, e)
        except Exception as e:
            logger.warning("Unexpected error in format detection for %s: %s", filename, e)

        return None

    def scan(self, filehandle, filename, fileformat, filename_ex):
        """Scan for malware.

        Args:
            filehandle: File data (memory mapped)
            filename: Path to file
            fileformat: Format info from format() method
            filename_ex: Extended filename info

        Returns:
            Tuple of (found, malware_name, malware_id, result)
        """
        try:
            if "ff_pdf" in fileformat:
                mm = filehandle

                if self.p_pdf_trojan_js and self.p_pdf_trojan_js.search(mm):
                    return True, "Trojan.PDF.Generic", 0, kernel.INFECTED

        except (IOError, OSError) as e:
            logger.debug("Scan IO error for %s: %s", filename, e)
        except Exception as e:
            logger.warning("Unexpected error scanning %s: %s", filename, e)

        return False, "", -1, kernel.NOT_FOUND

    def disinfect(self, filename, malware_id):
        """Disinfect malware.

        Args:
            filename: Path to infected file
            malware_id: Malware ID to disinfect

        Returns:
            True if successful, False otherwise
        """
        try:
            if malware_id == 0:
                # CWE-73: Safe file deletion
                filename_dir = os.path.dirname(filename) or os.getcwd()
                k2security.safe_remove_file(filename, filename_dir)
                return True

        except (IOError, OSError, k2security.SecurityError) as e:
            logger.debug("Disinfect error for %s: %s", filename, e)
        except Exception as e:
            logger.warning("Unexpected error disinfecting %s: %s", filename, e)

        return False

    def arclist(self, filename, fileformat, password=None):
        """List files in the archive (PDF objects).

        Args:
            filename: Path to PDF file
            fileformat: Format info from format() method

        Returns:
            List of [engine_id, filename] pairs
        """
        file_scan_list = []

        if "ff_pdf" not in fileformat:
            return file_scan_list

        try:
            zfile = self.__get_handle(filename)
            if zfile is None:
                return file_scan_list

            file_scan_list.extend(["arc_pdf", name] for name in zfile.namelist())

        except (IOError, OSError) as e:
            logger.debug("Archive list IO error for %s: %s", filename, e)
        except Exception as e:
            logger.warning("Unexpected error listing archive %s: %s", filename, e)

        return file_scan_list

    def unarc(self, arc_engine_id, arc_name, fname_in_arc):
        """Extract a file from the archive (PDF object).

        Args:
            arc_engine_id: Engine ID ('arc_pdf')
            arc_name: Path to PDF file
            fname_in_arc: Name of object to extract

        Returns:
            Extracted object data, or None on error
        """
        if arc_engine_id != "arc_pdf":
            return None

        try:
            zfile = self.__get_handle(arc_name)
            if zfile is None:
                return None

            return zfile.read(fname_in_arc)

        except (IOError, OSError) as e:
            logger.debug("Archive extract error for %s in %s: %s", fname_in_arc, arc_name, e)
        except Exception as e:
            logger.warning("Unexpected error extracting %s from %s: %s", fname_in_arc, arc_name, e)

        return None

    def arcclose(self):
        """Close all open archive handles."""
        for fname in list(self.handle.keys()):
            try:
                zfile = self.handle.get(fname)
                if zfile:
                    zfile.close()
            except (IOError, OSError) as e:
                logger.debug("Archive close IO error for %s: %s", fname, e)
            except Exception as e:
                logger.debug("Archive close error for %s: %s", fname, e)
            finally:
                self.handle.pop(fname, None)

    def mkarc(self, arc_engine_id, arc_name, file_infos):
        """Create an archive.

        Args:
            arc_engine_id: Engine ID ('arc_pdf')
            arc_name: Path to archive file
            file_infos: List of file info structures

        Returns:
            True if successful, False otherwise
        """
        # PDF cannot be recompressed, so it must be deleted
        return arc_engine_id == "arc_pdf"
